"""
Spatial random effects panel model based on: :cite:`KKP2007`
"""

__author__ = (
    "Luc Anselin anselin@uchicago.edu, Pedro Amaral pedroamaral@cedeplar.ufmg.br"
)

from scipy import sparse as SP
import numpy as np
from . import ols as OLS
from .utils import optim_moments, RegressionPropsY, get_spFilter, spdot, set_warn
from . import user_output as USER
from . import summary_output as SUMMARY
from . import regimes as REGI

# import warnings


__all__ = ["GM_KKP"]


class BaseGM_KKP(RegressionPropsY):

    '''
    Base GMM method for a spatial random effects panel model based on
    Kapoor, Kelejian and Prucha (2007) :cite:`KKP2007`.

    Parameters
    ----------
    y          : array
                 n*tx1 array for dependent variable
    x          : array
                 Two dimensional array with n*t rows and one column for each
                 independent (exogenous) variable
                 (note: must already include constant term)
    w          : spatial weights object
                 Spatial weights matrix
    full_weights: boolean
                  Considers different weights for each of the 6 moment
                  conditions if True or only 2 sets of weights for the
                  first 3 and the last 3 monent conditions if False (default)

    Attributes
    ----------
    betas        : array
                   kx1 array of estimated coefficients
    u            : array
                   nx1 array of residuals
    e_filtered   : array
                   nx1 array of spatially filtered residuals
    predy        : array
                   nx1 array of predicted y values
    n            : integer
                   Number of observations
    t            : integer
                   Number of time periods
    k            : integer
                   Number of variables for which coefficients are estimated
                   (including the constant)
    y            : array
                   nx1 array for dependent variable
    x            : array
                   Two dimensional array with n rows and one column for each
                   independent (exogenous) variable, including the constant
    vm           : array
                   Variance covariance matrix (kxk)
    """
    '''

    def __init__(self, y, x, w, full_weights=False):

        # 1a. OLS --> \tilde{\delta}
        ols = OLS.BaseOLS(y=y, x=x)
        self.x, self.y, self.n, self.k, self.xtx = ols.x, ols.y, ols.n, ols.k, ols.xtx
        N = w.n
        T = y.shape[0] // N
        moments, trace_w2 = _moments_kkp(w.sparse, ols.u, 0)
        lambda1, sig_v = optim_moments(moments, all_par=True)
        Tw = SP.kron(SP.identity(T), w.sparse)
        ub = Tw.dot(ols.u)
        ulu = ols.u - lambda1 * ub
        Q1 = SP.kron(np.ones((T, T)) / T, SP.identity(N))
        sig_1 = float(np.dot(ulu.T, Q1.dot(ulu)) / N)
        # print('initial_lamb_sig:',lambda1,sig_v,sig_1)
        # print('theta:', 1 - np.sqrt(sig_v)/ np.sqrt(sig_1))
        Xi_a = SP.diags([(sig_v * sig_v) / (T - 1), sig_1 * sig_1])
        if full_weights:
            Tau = _get_Tau(w.sparse, trace_w2)
        else:
            Tau = SP.identity(3)
        Xi = SP.kron(Xi_a, Tau)
        moments_b, _ = _moments_kkp(w.sparse, ols.u, 1, trace_w2)
        G = np.vstack((np.hstack((moments[0], np.zeros((3, 1)))), moments_b[0]))
        moments6 = [G, np.vstack((moments[1], moments_b[1]))]
        lambda2, sig_vb, sig_1b = optim_moments(
            moments6, vcX=Xi.toarray(), all_par=True, start=[lambda1, sig_v, sig_1]
        )
        # 2a. reg -->\hat{betas}
        theta = 1 - np.sqrt(sig_vb) / np.sqrt(sig_1b)
        # print('theta:', theta)
        gls_w = SP.identity(N * T) - theta * Q1

        # With omega
        xs = gls_w.dot(get_spFilter(w, lambda2, x))
        ys = gls_w.dot(get_spFilter(w, lambda2, y))
        ols_s = OLS.BaseOLS(y=ys, x=xs)
        self.predy = spdot(self.x, ols_s.betas)
        self.u = self.y - self.predy
        self.vm = ols_s.vm  # Check
        self.betas = np.vstack((ols_s.betas, lambda2, sig_vb, sig_1b))
        self.e_filtered = self.u - lambda2 * SP.kron(SP.identity(T), w.sparse).dot(
            self.u
        )
        self.t, self.n = T, N
        self._cache = {}


class GM_KKP(BaseGM_KKP, REGI.Regimes_Frame):

    '''
    GMM method for a spatial random effects panel model based on
    Kapoor, Kelejian and Prucha (2007) :cite:`KKP2007`.

    Parameters
    ----------
    y          : array
                 n*tx1 or nxt array for dependent variable
    x          : array
                 Two dimensional array with n*t rows and k columns for
                 independent (exogenous) variable or n rows and k*t columns
                 (note, must not include a constant term)
    w          : spatial weights object
                 Spatial weights matrix, nxn
    full_weights: boolean
                  Considers different weights for each of the 6 moment
                  conditions if True or only 2 sets of weights for the
                  first 3 and the last 3 moment conditions if False (default)
    regimes      : list
                   List of n values with the mapping of each
                   observation to a regime. Assumed to be aligned with 'y'.
    vm           : boolean
                   If True, include variance-covariance matrix in summary
                   results
    name_y       : string or list of strings
                   Name of dependent variable for use in output
    name_x       : list of strings
                   Names of independent variables for use in output
    name_w       : string
                   Name of weights matrix for use in output
    name_ds      : string
                   Name of dataset for use in output
    name_regimes : string
                   Name of regime variable for use in the output

    Attributes
    ----------
    betas        : array
                   kx1 array of estimated coefficients
    u            : array
                   nx1 array of residuals
    e_filtered   : array
                   nx1 array of spatially filtered residuals
    predy        : array
                   nx1 array of predicted y values
    n            : integer
                   Number of observations
    t            : integer
                   Number of time periods
    k            : integer
                   Number of variables for which coefficients are estimated
                   (including the constant)
    y            : array
                   nx1 array for dependent variable
    x            : array
                   Two dimensional array with n rows and one column for each
                   independent (exogenous) variable, including the constant
    vm           : array
                   Variance covariance matrix (kxk)
    chow         : tuple
                   Contains 2 elements. 1: Pair of Wald statistic and p-value
                   for the setup of global regime stability. 2: array with Wald
                   statistic (col 0) and its p-value (col 1) for each beta that
                   varies across regimes.
                   Exists only if regimes is not None.
    name_y       : string
                   Name of dependent variable for use in output
    name_x       : list of strings
                   Names of independent variables for use in output
    name_w       : string
                   Name of weights matrix for use in output
    name_ds      : string
                   Name of dataset for use in output
    name_regimes : string
                   Name of regime variable for use in the output
    title        : string
                   Name of the regression method used
    """
    Examples
    --------
    We first need to import the needed modules, namely numpy to convert the
    data we read into arrays that ``spreg`` understands and ``pysal`` to
    perform all the analysis.
    >>> from spreg import GM_KKP
    >>> import numpy as np
    >>> import libpysal
    Open data on NCOVR US County Homicides (3085 areas) using libpysal.io.open().
    This is the DBF associated with the NAT shapefile. Note that
    libpysal.io.open() also reads data in CSV format; The GM_KKP function requires
    data to be passed in as numpy arrays, hence the user can read their
    data in using any method.
    >>> nat = libpysal.examples.load_example('NCOVR')
    >>> db = libpysal.io.open(nat.get_path("NAT.dbf"),'r')
    Extract the HR (homicide rates) data in the 70's, 80's and 90's from the DBF file
    and make it the dependent variable for the regression. Note that the data can also
    be passed in the long format instead of wide format (i.e. a vector with n*t rows
    and a single column for the dependent variable and a matrix of dimension n*txk
    for the independent variables).
    >>> name_y = ['HR70','HR80','HR90']
    >>> y = np.array([db.by_col(name) for name in name_y]).T
    Extract RD and PS in the same time periods from the DBF to be used as
    independent variables in the regression.  Note that PySAL requires this to
    be an nxk*t numpy array, where k is the number of independent variables (not
    including a constant) and t is the number of time periods. Data must be
    organized in a way that all time periods of a given variable are side-by-side
    and in the correct time order.
    By default a vector of ones will be added to the independent variables passed in.
    >>> name_x = ['RD70','RD80','RD90','PS70','PS80','PS90']
    >>> x = np.array([db.by_col(name) for name in name_x]).T
    Since we want to run a spatial error panel model, we need to specify the spatial
    weights matrix that includes the spatial configuration of the observations
    into the error component of the model. To do that, we can open an already
    existing gal file or create a new one. In this case, we will create one
    from ``NAT.shp``.
    >>> w = libpysal.weights.Queen.from_shapefile(libpysal.examples.get_path("NAT.shp"))
    Unless there is a good reason not to do it, the weights have to be
    row-standardized so every row of the matrix sums to one. Among other
    things, his allows to interpret the spatial lag of a variable as the
    average value of the neighboring observations. In PySAL, this can be
    easily performed in the following way:
    >>> w.transform = 'r'
    We are all set with the preliminaries, we are good to run the model. In this
    case, we will need the variables and the weights matrix. If we want to
    have the names of the variables printed in the output summary, we will
    have to pass them in as well, although this is optional. In this example
    we set full_weights to False (the default), indicating that we will use
    only 2 sets of moments weights for the first 3 and the last 3 moment conditions.
    >>> reg = GM_KKP(y,x,w,full_weights=False,name_y=name_y, name_x=name_x)
    Warning: Assuming time data is in wide format, i.e. y[0] refers to T0, y[1], refers to T1, etc.
     Similarly, assuming x[0:k] refers to independent variables for T0, x[k+1:2k] refers to T1, etc.
    Once we have run the model, we can explore a little bit the output. We can
    either request a printout of the results with the command print(reg.summary) or
    check out the individual attributes of GM_KKP:
    >>> print(reg.summary)
    REGRESSION
    ----------
    SUMMARY OF OUTPUT: GM SPATIAL ERROR PANEL MODEL - RANDOM EFFECTS (KKP)
    ----------------------------------------------------------------------
    Data set            :     unknown
    Weights matrix      :     unknown
    Dependent Variable  :          HR                Number of Observations:        3085
    Mean dependent var  :      6.4983                Number of Variables   :           3
    S.D. dependent var  :      6.9529                Degrees of Freedom    :        3082
    Pseudo R-squared    :      0.3248
    <BLANKLINE>
    ------------------------------------------------------------------------------------
                Variable     Coefficient       Std.Error     z-Statistic     Probability
    ------------------------------------------------------------------------------------
                CONSTANT       6.4922156       0.1126713      57.6208690       0.0000000
                      RD       3.6244575       0.0877475      41.3055536       0.0000000
                      PS       1.3118778       0.0852516      15.3883058       0.0000000
                  lambda       0.4177759
                sigma2_v      22.8190822
                sigma2_1      39.9099323
    ------------------------------------------------------------------------------------
    ================================ END OF REPORT =====================================
    >>> print(reg.name_x)
    ['CONSTANT', 'RD', 'PS', 'lambda', ' sigma2_v', 'sigma2_1']
    The attribute reg.betas contains all the coefficients: betas, the spatial error
    coefficient lambda, sig2_v and sig2_1:
    >>> print(np.around(reg.betas,4))
    [[ 6.4922]
     [ 3.6245]
     [ 1.3119]
     [ 0.4178]
     [22.8191]
     [39.9099]]
    Finally, we can check the standard erros of the betas:
    >>> print(np.around(np.sqrt(reg.vm.diagonal().reshape(3,1)),4))
    [[0.1127]
     [0.0877]
     [0.0853]]
    '''

    def __init__(
        self,
        y,
        x,
        w,
        full_weights=False,
        regimes=None,
        vm=False,
        name_y=None,
        name_x=None,
        name_w=None,
        name_ds=None,
        name_regimes=None,
    ):
        n_rows = USER.check_arrays(y, x)
        bigy, bigx, name_y, name_x = _get_panel_data(y, x, w, name_y, name_x)
        USER.check_weights(w, bigy, w_required=True, time=True)
        x_constant, name_x, warn = USER.check_constant(bigx, name_x)
        set_warn(self, warn)
        self.title = "GM SPATIAL ERROR PANEL MODEL - RANDOM EFFECTS (KKP)"
        self.name_x = USER.set_name_x(name_x, x_constant)

        if regimes is not None:
            self.regimes = regimes
            self.name_regimes = USER.set_name_ds(name_regimes)
            regimes_l = self._set_regimes(w, bigy.shape[0])
            self.name_x_r = self.name_x
            x_constant, self.name_x = REGI.Regimes_Frame.__init__(
                self,
                x_constant,
                regimes_l,
                constant_regi=False,
                cols2regi="all",
                names=self.name_x,
            )

        BaseGM_KKP.__init__(self, bigy, x_constant, w, full_weights=full_weights)
        self.name_ds = USER.set_name_ds(name_ds)
        self.name_y = USER.set_name_y(name_y)
        self.name_x.extend(["lambda", " sigma2_v", "sigma2_1"])
        self.name_w = USER.set_name_w(name_w, w)
        if regimes is not None:
            self.kf += 3
            self.chow = REGI.Chow(self)
            self.title += " WITH REGIMES"
            regimes = True
        SUMMARY.GM_Panels(reg=self, w=w, vm=vm, regimes=regimes)

    def _set_regimes(self, w, n_rows):  # Must add case for regime_err_sep = True
        self.constant_regi = "many"
        self.cols2regi = "all"
        self.regime_err_sep = False
        self.regimes_set = REGI._get_regimes_set(self.regimes)
        if len(self.regimes) == w.n:
            regimes_l = self.regimes * (n_rows // w.n)
        elif len(self.regimes) == n_rows:
            regimes_l = self.regimes
        else:
            raise Exception("The lenght of 'regimes' must be either equal to n or n*t.")
        return regimes_l


def _moments_kkp(ws, u, i, trace_w2=None):
    """
    Compute G and g matrices for the KKP model.
    ...

    Parameters
    ----------
    ws          : Sparse matrix
                  Spatial weights sparse matrix
    u           : array
                  Residuals. nx1 array assumed to be aligned with w

    i       : integer
                  0 if Q0, 1 if Q1
    trace_w2    : float
                  trace of WW. Computed in 1st step and saved for step 2.
    Returns
    -------
    moments     : list
                  List of two arrays corresponding to the matrices 'G' and
                  'g', respectively.
    trace_w2    : float
                  trace of WW. Computed in 1st step and saved for step 2.
    """
    N = ws.shape[0]
    T = u.shape[0] // N
    if i == 0:
        Q = SP.kron(SP.identity(T) - np.ones((T, T)) / T, SP.identity(N))
    else:
        Q = SP.kron(np.ones((T, T)) / T, SP.identity(N))
    Tw = SP.kron(SP.identity(T), ws)
    ub = Tw.dot(u)
    ubb = Tw.dot(ub)
    Qu = Q.dot(u)
    Qub = Q.dot(ub)
    Qubb = Q.dot(ubb)
    G11 = float(2 * np.dot(u.T, Qub))
    G12 = float(-np.dot(ub.T, Qub))
    G21 = float(2 * np.dot(ubb.T, Qub))
    G22 = float(-np.dot(ubb.T, Qubb))
    G31 = float(np.dot(u.T, Qubb) + np.dot(ub.T, Qub))
    G32 = float(-np.dot(ub.T, Qubb))
    if trace_w2 == None:
        trace_w2 = (ws.power(2)).sum()
    G23 = ((T - 1) ** (1 - i)) * trace_w2
    if i == 0:
        G = np.array(
            [[G11, G12, N * (T - 1) ** (1 - i)], [G21, G22, G23], [G31, G32, 0]]
        ) / (N * (T - 1) ** (1 - i))
    else:
        G = np.array(
            [
                [G11, G12, 0, N * (T - 1) ** (1 - i)],
                [G21, G22, 0, G23],
                [G31, G32, 0, 0],
            ]
        ) / (N * (T - 1) ** (1 - i))
    g1 = float(np.dot(u.T, Qu))
    g2 = float(np.dot(ub.T, Qub))
    g3 = float(np.dot(u.T, Qub))
    g = np.array([[g1, g2, g3]]).T / (N * (T - 1) ** (1 - i))
    return [G, g], trace_w2


def _get_Tau(ws, trace_w2):
    """
    Computes Tau as in :cite:`KKP2007`.
    ...

    Parameters
    ----------
    ws          : Sparse matrix
                  Spatial weights sparse matrix
    trace_w2    : float
                  trace of WW. Computed in 1st step of _moments_kkp
    """
    N = ws.shape[0]
    T12 = 2 * trace_w2 / N
    wtw = ws.T.dot(ws)
    T22 = wtw.power(2).sum()
    wtpw = ws.T + ws
    T23 = wtw.multiply(wtpw).sum()
    d_wwpwtw = ws.multiply(ws.T).sum(0) + wtw.diagonal()
    T33 = d_wwpwtw.sum()
    Tau = np.array([[2 * N, T12, 0], [T12, T22, T23], [0, T23, T33]]) / N
    return Tau


def _get_panel_data(y, x, w, name_y, name_x):
    """
    Performs some checks on the data structure and converts from wide to long if needed.
    ...

    Parameters
    ----------
    y          : array
                 n*tx1 or nxt array for dependent variable
    x          : array
                 Two dimensional array with n*t rows and k columns for
                 independent (exogenous) variable or n rows and k*t columns
                 (note, must not include a constant term)
    name_y       : string or list of strings
                   Name of dependent variable for use in output
    name_x       : list of strings
                   Names of independent variables for use in output
    """

    if y.shape[0] / w.n != y.shape[0] // w.n:
        raise Exception("y must be ntx1 or nxt, and w must be an nxn PySAL W object.")
    N, T = y.shape[0], y.shape[1]
    k = x.shape[1] // T
    if x.shape[0] != N and x.shape[0] != N * T:
        raise Exception(
            "X must have either n rows and k*t columns or n*t rows and k columns."
        )
    if x.shape[1] != k and x.shape[1] != k * T:
        raise Exception(
            "X must have either n rows and k*t columns or n*t rows and k columns."
        )
    if y.shape[1] > 1:
        message = (
            "Assuming time data is in wide format, i.e. y[0] refers to T0, y[1], refers to T1, etc."
            "\n Similarly, assuming x[0:k] refers to independent variables for T0, x[k+1:2k] refers to T1, etc."
        )
        print("Warning: " + message)
        # warnings.warn(message)

        if y.shape[1] != T:
            raise Exception(
                "y in wide format must have t columns and be compatible with x's k*t columns."
            )

        bigy = y.reshape((y.size, 1), order="F")

        bigx = x[:, 0:T].reshape((N * T, 1), order="F")
        for i in range(1, k):
            bigx = np.hstack(
                (bigx, x[:, T * i : T * (i + 1)].reshape((N * T, 1), order="F"))
            )
    else:
        bigy, bigx = y, x

    if name_y:
        if not isinstance(name_y, str) and not isinstance(name_y, list):
            raise Exception("name_y must either be strings or a list of strings.")
        if len(name_y) > 1 and isinstance(name_y, list):
            name_y = "".join([i for i in name_y[0] if not i.isdigit()])
        if len(name_y) == 1 and isinstance(name_y, list):
            name_y = name_y[0]
    if name_x:
        if len(name_x) != k * T and len(name_x) != k:
            raise Exception(
                "Names of columns in X must have exactly either k or k*t elements."
            )
        if len(name_x) > k:
            name_bigx = []
            for i in range(k):
                name_bigx.append("".join([j for j in name_x[i * T] if not j.isdigit()]))
            name_x = name_bigx

    return bigy, bigx, name_y, name_x


def _test():
    import doctest

    start_suppress = np.get_printoptions()["suppress"]
    np.set_printoptions(suppress=True)
    doctest.testmod()
    np.set_printoptions(suppress=start_suppress)


if __name__ == "__main__":
    _test()



import unittest
import numpy as np
import libpysal
import spreg as EC
from scipy import sparse
from libpysal.common import RTOL

PEGP = libpysal.examples.get_path


class TestBaseOLS(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(PEGP("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        X.append(db.by_col("CRIME"))
        self.X = np.array(X).T
        self.w = libpysal.weights.Rook.from_shapefile(PEGP("columbus.shp"))

    def test_ols(self):
        self.X = np.hstack((np.ones(self.y.shape), self.X))
        self.X = sparse.csr_matrix(self.X)
        ols = EC.ols.BaseOLS(self.y, self.X)
        np.testing.assert_allclose(
            ols.betas, np.array([[46.42818268], [0.62898397], [-0.48488854]])
        )
        vm = np.array(
            [
                [1.74022453e02, -6.52060364e00, -2.15109867e00],
                [-6.52060364e00, 2.87200008e-01, 6.80956787e-02],
                [-2.15109867e00, 6.80956787e-02, 3.33693910e-02],
            ]
        )
        np.testing.assert_allclose(ols.vm, vm, RTOL)

    def test_OLS(self):
        self.X = sparse.csr_matrix(self.X)
        ols = EC.OLS(
            self.y,
            self.X,
            self.w,
            spat_diag=True,
            moran=True,
            name_y="home value",
            name_x=["income", "crime"],
            name_ds="columbus",
            nonspat_diag=True,
            white_test=True,
        )

        np.testing.assert_allclose(ols.aic, 408.73548964604873, RTOL)
        np.testing.assert_allclose(ols.ar2, 0.32123239427957662, RTOL)
        np.testing.assert_allclose(
            ols.betas, np.array([[46.42818268], [0.62898397], [-0.48488854]]), RTOL
        )
        bp = np.array([2, 5.7667905131212587, 0.05594449410070558])
        ols_bp = np.array(
            [
                ols.breusch_pagan["df"],
                ols.breusch_pagan["bp"],
                ols.breusch_pagan["pvalue"],
            ]
        )
        np.testing.assert_allclose(bp, ols_bp, RTOL)
        np.testing.assert_allclose(
            ols.f_stat, (12.358198885356581, 5.0636903313953024e-05), RTOL
        )
        jb = np.array([2, 39.706155069114878, 2.387360356860208e-09])
        ols_jb = np.array(
            [ols.jarque_bera["df"], ols.jarque_bera["jb"], ols.jarque_bera["pvalue"]]
        )
        np.testing.assert_allclose(ols_jb, jb, RTOL)
        white = np.array([5, 2.90606708, 0.71446484])
        ols_white = np.array([ols.white["df"], ols.white["wh"], ols.white["pvalue"]])
        np.testing.assert_allclose(ols_white, white, RTOL)
        np.testing.assert_equal(ols.k, 3)
        kb = {"df": 2, "kb": 2.2700383871478675, "pvalue": 0.32141595215434604}
        for key in kb:
            np.testing.assert_allclose(ols.koenker_bassett[key], kb[key], RTOL)
        np.testing.assert_allclose(
            ols.lm_error, (4.1508117035117893, 0.041614570655392716), RTOL
        )
        np.testing.assert_allclose(
            ols.lm_lag, (0.98279980617162233, 0.32150855529063727), RTOL
        )
        np.testing.assert_allclose(
            ols.lm_sarma, (4.3222725729143736, 0.11519415308749938), RTOL
        )
        np.testing.assert_allclose(ols.logll, -201.3677448230244, RTOL)
        np.testing.assert_allclose(ols.mean_y, 38.436224469387746, RTOL)
        np.testing.assert_allclose(ols.moran_res[0], 0.20373540938, RTOL)
        np.testing.assert_allclose(ols.moran_res[1], 2.59180452208, RTOL)
        np.testing.assert_allclose(ols.moran_res[2], 0.00954740031251, RTOL)
        np.testing.assert_allclose(ols.mulColli, 12.537554873824675, RTOL)
        np.testing.assert_equal(ols.n, 49)
        np.testing.assert_equal(ols.name_ds, "columbus")
        np.testing.assert_equal(ols.name_gwk, None)
        np.testing.assert_equal(ols.name_w, "unknown")
        np.testing.assert_equal(ols.name_x, ["CONSTANT", "income", "crime"])
        np.testing.assert_equal(ols.name_y, "home value")
        np.testing.assert_allclose(ols.predy[3], np.array([33.53969014]), RTOL)
        np.testing.assert_allclose(ols.r2, 0.34951437785126105, RTOL)
        np.testing.assert_allclose(
            ols.rlm_error, (3.3394727667427513, 0.067636278225568919), RTOL
        )
        np.testing.assert_allclose(
            ols.rlm_lag, (0.17146086940258459, 0.67881673703455414), RTOL
        )
        np.testing.assert_equal(ols.robust, "unadjusted")
        np.testing.assert_allclose(ols.schwarz, 414.41095054038061, RTOL)
        np.testing.assert_allclose(ols.sig2, 231.4568494392652, RTOL)
        np.testing.assert_allclose(ols.sig2ML, 217.28602192257551, RTOL)
        np.testing.assert_allclose(ols.sig2n, 217.28602192257551, RTOL)

        np.testing.assert_allclose(ols.t_stat[2][0], -2.65440864272, RTOL)
        np.testing.assert_allclose(ols.t_stat[2][1], 0.0108745049098, RTOL)


if __name__ == "__main__":
    unittest.main()



"""
Utilities for panel data estimation
"""

__author__ = "Wei Kang weikang9009@gmail.com, \
              Pedro Amaral pedroamaral@cedeplar.ufmg.br, \
              Pablo Estrada pabloestradace@gmail.com"

import numpy as np
from scipy import sparse as sp
from .sputils import spdot

__all__ = ["check_panel", "demean_panel"]


def check_panel(y, x, w, name_y, name_x):
    """
    Check the data structure and converts from wide to long if needed.

    Parameters
    ----------
    y          : array
                 n*tx1 or nxt array for dependent variable
    x          : array
                 Two dimensional array with n*t rows and k columns for
                 independent (exogenous) variable or n rows and k*t columns
                 (note, must not include a constant term)
    name_y     : string or list of strings
                 Name of dependent variable for use in output
    name_x     : list of strings
                 Names of independent variables for use in output
    """

    # Check if 'y' is a balanced panel with respect to 'W'
    if y.shape[0] / w.n != y.shape[0] // w.n:
        raise Exception("y must be ntx1 or nxt, and w must be an nxn PySAL W" "object.")
    # Wide format
    if y.shape[1] > 1:
        warn = (
            "Assuming panel is in wide format.\n"
            "y[:, 0] refers to T0, y[:, 1] refers to T1, etc.\n"
            "x[:, 0:T] refers to T periods of k1, x[:, T+1:2T] refers "
            "to k2, etc."
        )
        N, T = y.shape[0], y.shape[1]
        k = x.shape[1] // T
        bigy = y.reshape((y.size, 1), order="F")
        bigx = x[:, 0:T].reshape((N * T, 1), order="F")
        for i in range(1, k):
            bigx = np.hstack(
                (bigx, x[:, T * i : T * (i + 1)].reshape((N * T, 1), order="F"))
            )
    # Long format
    else:
        warn = (
            "Assuming panel is in long format.\n"
            "y[0:N] refers to T0, y[N+1:2N] refers to T1, etc.\n"
            "x[0:N] refers to T0, x[N+1:2N] refers to T1, etc."
        )
        T = y.shape[0] // w.n
        N = w.n
        k = x.shape[1]
        bigy, bigx = y, x
    # Fix name_y and name_x
    if name_y:
        if not isinstance(name_y, str) and not isinstance(name_y, list):
            raise Exception("name_y must either be strings or a list of" "strings.")
        if len(name_y) > 1 and isinstance(name_y, list):
            name_y = "".join([i for i in name_y[0] if not i.isdigit()])
        if len(name_y) == 1 and isinstance(name_y, list):
            name_y = name_y[0]
    if name_x:
        if len(name_x) != k * T and len(name_x) != k:
            raise Exception(
                "Names of columns in X must have exactly either" "k or k*t elements."
            )
        if len(name_x) > k:
            name_bigx = []
            for i in range(k):
                namek = "".join([j for j in name_x[i * T] if not j.isdigit()])
                name_bigx.append(namek)
            name_x = name_bigx

    return bigy, bigx, name_y, name_x, warn


def demean_panel(arr, n, t, phi=0):
    """
    Returns demeaned variable.

    Parameters
    ----------
    arr         : array
                  n*tx1 array
    n           : integer
                  Number of observations
    t           : integer
                  Number of time periods
    phi         : float
                  Weight from 0 to 1 attached to the cross-sectional component
                  of the data. If phi=0, then it is the demeaning procedure.
                  If phi=1, then the data doesn't change at all.

    Returns
    -------
    arr_dm      : array
                  Demeaned variable
    """

    one = np.ones((t, 1))
    J = sp.identity(t) - (1 - phi) * (1 / t) * spdot(one, one.T)
    Q = sp.kron(J, sp.identity(n), format="csr")
    arr_dm = spdot(Q, arr)

    return arr_dm



"""
Diagnostics for SUR and 3SLS estimation
"""

__author__ = "Luc Anselin lanselin@gmail.com,    \
             Pedro V. Amaral pedrovma@gmail.com  \
             Tony Aburaad taburaad@uchicago.edu"


import numpy as np
import scipy.stats as stats
import numpy.linalg as la
from .sur_utils import sur_dict2mat, sur_mat2dict, sur_corr, spdot
from .regimes import buildR1var, wald_test


__all__ = ["sur_setp", "sur_lrtest", "sur_lmtest", "lam_setp", "surLMe", "surLMlag"]


def sur_setp(bigB, varb):
    """
    Utility to compute standard error, t and p-value

    Parameters
    ----------
    bigB    : dictionary
              of regression coefficient estimates,
              one vector by equation
    varb    : array
              variance-covariance matrix of coefficients

    Returns
    -------
    surinfdict : dictionary
                 with standard error, t-value, and
                 p-value array, one for each equation

    """
    vvb = varb.diagonal()
    n_eq = len(bigB.keys())
    bigK = np.zeros((n_eq, 1), dtype=np.int_)
    for r in range(n_eq):
        bigK[r] = bigB[r].shape[0]
    b = sur_dict2mat(bigB)
    se = np.sqrt(vvb)
    se.resize(len(se), 1)
    t = np.divide(b, se)
    tp = stats.norm.sf(abs(t)) * 2
    surinf = np.hstack((se, t, tp))
    surinfdict = sur_mat2dict(surinf, bigK)
    return surinfdict


def lam_setp(lam, vm):
    """
    Standard errors, t-test and p-value for lambda in SUR Error ML

    Parameters
    ----------
    lam        : array
                 n_eq x 1 array with ML estimates for spatial error
                 autoregressive coefficient
    vm         : array
                 n_eq x n_eq subset of variance-covariance matrix for
                 lambda and Sigma in SUR Error ML
                 (needs to be subset from full vm)

    Returns
    -------
               : tuple
                 with arrays for standard error, t-value and p-value
                 (each element in the tuple is an n_eq x 1 array)

    """
    vvb = vm.diagonal()
    se = np.sqrt(vvb)
    se.resize(len(se), 1)
    t = np.divide(lam, se)
    tp = stats.norm.sf(abs(t)) * 2
    return (se, t, tp)


def sur_lrtest(n, n_eq, ldetS0, ldetS1):
    """
    Likelihood Ratio test on off-diagonal elements of Sigma

    Parameters
    ----------
    n        : int
               cross-sectional dimension (number of observations for an equation)
    n_eq     : int
               number of equations
    ldetS0   : float
               log determinant of Sigma for OLS case
    ldetS1   : float
               log determinant of Sigma for SUR case (should be iterated)

    Returns
    -------
    (lrtest,M,pvalue) : tuple
                        with value of test statistic (lrtest),
                        degrees of freedom (M, as an integer)
                        p-value

    """
    M = n_eq * (n_eq - 1) / 2.0
    lrtest = n * (ldetS0 - ldetS1)
    pvalue = stats.chi2.sf(lrtest, M)
    return (lrtest, int(M), pvalue)


def sur_lmtest(n, n_eq, sig):
    """
    Lagrange Multiplier test on off-diagonal elements of Sigma

    Parameters
    ----------
    n        : int
               cross-sectional dimension (number of observations for an equation)
    n_eq     : int
               number of equations
    sig      : array
               inter-equation covariance matrix for null model (OLS)

    Returns
    -------
    (lmtest,M,pvalue) : tuple
                        with value of test statistic (lmtest),
                        degrees of freedom (M, as an integer)
                        p-value
    """
    R = sur_corr(sig)
    tr = np.trace(np.dot(R.T, R))
    M = n_eq * (n_eq - 1) / 2.0
    lmtest = (n / 2.0) * (tr - n_eq)
    pvalue = stats.chi2.sf(lmtest, M)
    return (lmtest, int(M), pvalue)


def surLMe(n_eq, WS, bigE, sig):
    """
    Lagrange Multiplier test on error spatial autocorrelation in SUR

    Parameters
    ----------
    n_eq       : int
                 number of equations
    WS         : array
                 spatial weights matrix in sparse form
    bigE       : array
                 n x n_eq matrix of residuals by equation
    sig        : array
                 cross-equation error covariance matrix

    Returns
    -------
    (LMe,n_eq,pvalue) : tuple
                        with value of statistic (LMe), degrees
                        of freedom (n_eq) and p-value

    """
    # spatially lagged residuals
    WbigE = WS * bigE
    # score
    EWE = np.dot(bigE.T, WbigE)
    sigi = la.inv(sig)
    SEWE = sigi * EWE
    # score = SEWE.sum(axis=1)
    # score.resize(n_eq,1)
    # note score is column sum of Sig_i * E'WE, a 1 by n_eq row vector
    # previously stored as column
    score = SEWE.sum(axis=0)
    score.resize(1, n_eq)

    # trace terms
    WW = WS * WS
    trWW = np.sum(WW.diagonal())
    WTW = WS.T * WS
    trWtW = np.sum(WTW.diagonal())
    # denominator
    SiS = sigi * sig
    Tii = trWW * np.identity(n_eq)
    tSiS = trWtW * SiS
    denom = Tii + tSiS
    idenom = la.inv(denom)
    # test statistic
    # LMe = np.dot(np.dot(score.T,idenom),score)[0][0]
    # score is now row vector
    LMe = np.dot(np.dot(score, idenom), score.T)[0][0]
    pvalue = stats.chi2.sf(LMe, n_eq)
    return (LMe, n_eq, pvalue)


def surLMlag(n_eq, WS, bigy, bigX, bigE, bigYP, sig, varb):
    """
    Lagrange Multiplier test on lag spatial autocorrelation in SUR

    Parameters
    ----------
    n_eq       : int
                 number of equations
    WS         : spatial weights matrix in sparse form
    bigy       : dictionary
                 with y values
    bigX       : dictionary
                 with X values
    bigE       : array
                 n x n_eq matrix of residuals by equation
    bigYP      : array
                 n x n_eq matrix of predicted values by equation
    sig        : array
                 cross-equation error covariance matrix
    varb       : array
                 variance-covariance matrix for b coefficients (inverse of Ibb)

    Returns
    -------
    (LMlag,n_eq,pvalue) : tuple
                          with value of statistic (LMlag), degrees
                          of freedom (n_eq) and p-value

    """
    # Score
    Y = np.hstack([bigy[r] for r in range(n_eq)])
    WY = WS * Y
    EWY = np.dot(bigE.T, WY)
    sigi = la.inv(sig)
    SEWE = sigi * EWY
    score = SEWE.sum(axis=0)  # column sums
    score.resize(1, n_eq)  # score as a row vector

    # I(rho,rho) as partitioned inverse, eq 72
    # trace terms
    WW = WS * WS
    trWW = np.sum(WW.diagonal())  # T1
    WTW = WS.T * WS
    trWtW = np.sum(WTW.diagonal())  # T2

    # I(rho,rho)
    SiS = sigi * sig
    Tii = trWW * np.identity(n_eq)  # T1It
    tSiS = trWtW * SiS
    firstHalf = Tii + tSiS
    WbigYP = WS * bigYP
    inner = np.dot(WbigYP.T, WbigYP)
    secondHalf = sigi * inner
    Ipp = firstHalf + secondHalf  # eq. 75

    # I(b,b) inverse is varb

    # I(b,rho)
    bp = sigi[0,] * spdot(
        bigX[0].T, WbigYP
    )  # initialize
    for r in range(1, n_eq):
        bpwork = (
            sigi[
                r,
            ]
            * spdot(bigX[r].T, WbigYP)
        )
        bp = np.vstack((bp, bpwork))
    # partitioned part
    i_inner = Ipp - np.dot(np.dot(bp.T, varb), bp)
    # partitioned inverse of information matrix
    Ippi = la.inv(i_inner)

    # test statistic
    LMlag = np.dot(np.dot(score, Ippi), score.T)[0][0]
    # p-value
    pvalue = stats.chi2.sf(LMlag, n_eq)
    return (LMlag, n_eq, pvalue)


def sur_chow(n_eq, bigK, bSUR, varb):
    """
    test on constancy of regression coefficients across equations in
    a SUR specification

    Note: requires a previous check on constancy of number of coefficients
    across equations; no other checks are carried out, so it is possible
    that the results are meaningless if the variables are not listed in
    the same order in each equation.

    Parameters
    ----------
    n_eq       : int
                 number of equations
    bigK       : array
                 with the number of variables by equation (includes constant)
    bSUR       : dictionary
                 with the SUR regression coefficients by equation
    varb       : array
                 the variance-covariance matrix for the SUR regression
                 coefficients

    Returns
    -------
    test       : array
                 a list with for each coefficient (in order) a tuple with the
                 value of the test statistic, the degrees of freedom, and the
                 p-value

    """
    kr = bigK[0][0]
    test = []
    bb = sur_dict2mat(bSUR)
    kf = 0
    nr = n_eq
    df = n_eq - 1
    for i in range(kr):
        Ri = buildR1var(i, kr, kf, 0, nr)
        tt, p = wald_test(bb, Ri, np.zeros((df, 1)), varb)
        test.append((tt, df, p))
    return test


def sur_joinrho(n_eq, bigK, bSUR, varb):
    """
    Test on joint significance of spatial autoregressive coefficient in SUR

    Parameters
    ----------
    n_eq       : int
                 number of equations
    bigK       : array
                 n_eq x 1 array with number of variables by equation
                 (includes constant term, exogenous and endogeneous and
                 spatial lag)
    bSUR       : dictionary
                 with regression coefficients by equation, with
                 the spatial autoregressive term as last
    varb       : array
                 variance-covariance matrix for regression coefficients

    Returns
    -------
               : tuple
                 with test statistic, degrees of freedom, p-value

    """
    bb = sur_dict2mat(bSUR)
    R = np.zeros((n_eq, varb.shape[0]))
    q = np.zeros((n_eq, 1))
    kc = -1
    for i in range(n_eq):
        kc = kc + bigK[i]
        R[i, kc] = 1
    w, p = wald_test(bb, R, q, varb)
    return (w, n_eq, p)



import unittest
import libpysal
import numpy as np
from scipy import sparse
from spreg.ml_error_regimes import ML_Error_Regimes
from spreg.ml_error import ML_Error
from libpysal.common import RTOL
from warnings import filterwarnings

filterwarnings("ignore", category=sparse.SparseEfficiencyWarning)


class TestMLError(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("baltim.dbf"), "r")
        self.ds_name = "baltim.dbf"
        self.y_name = "PRICE"
        self.y = np.array(db.by_col(self.y_name)).T
        self.y.shape = (len(self.y), 1)
        self.x_names = ["NROOM", "AGE", "SQFT"]
        self.x = np.array([db.by_col(var) for var in self.x_names]).T
        ww = libpysal.io.open(libpysal.examples.get_path("baltim_q.gal"))
        self.w = ww.read()
        ww.close()
        self.w_name = "baltim_q.gal"
        self.w.transform = "r"
        self.regimes = db.by_col("CITCOU")
        # Artficial:
        n = 256
        self.n2 = int(n / 2)
        self.x_a1 = np.random.uniform(-10, 10, (n, 1))
        self.x_a2 = np.random.uniform(1, 5, (n, 1))
        self.q_a = self.x_a2 + np.random.normal(0, 1, (n, 1))
        self.x_a = np.hstack((self.x_a1, self.x_a2))
        self.y_a = np.dot(
            np.hstack((np.ones((n, 1)), self.x_a)), np.array([[1], [0.5], [2]])
        ) + np.random.normal(0, 1, (n, 1))
        latt = int(np.sqrt(n))
        self.w_a = libpysal.weights.util.lat2W(latt, latt)
        self.w_a.transform = "r"
        self.regi_a = [0] * (n // 2) + [1] * (n // 2)  # must be floor!
        self.w_a1 = libpysal.weights.util.lat2W(latt // 2, latt)
        self.w_a1.transform = "r"

    def test_model1(self):
        reg = ML_Error_Regimes(
            self.y,
            self.x,
            self.regimes,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_w=self.w_name,
            name_ds=self.ds_name,
            name_regimes="CITCOU",
            regime_err_sep=False,
        )
        betas = np.array(
            [
                [-2.39491278],
                [4.873757],
                [-0.02911854],
                [0.33275008],
                [31.79618475],
                [2.98102401],
                [-0.23710892],
                [0.80581127],
                [0.61770744],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([30.46599009])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([16.53400991])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 211
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 8
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([47.0])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, 4.0, 148.0, 11.25, 0.0, 0.0, 0.0, 0.0])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        e = np.array([34.69181334])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        my = 44.307180094786695
        np.testing.assert_allclose(reg.mean_y, my, RTOL)
        sy = 23.606076835380495
        np.testing.assert_allclose(reg.std_y, sy, RTOL)
        vm = np.array(
            [
                58.50551173,
                2.42952002,
                0.00721525,
                0.06391736,
                80.59249161,
                3.1610047,
                0.0119782,
                0.0499432,
                0.00502785,
            ]
        )
        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL)
        sig2 = np.array([[209.60639741]])
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.43600837301477025
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        std_err = np.array(
            [
                7.64888957,
                1.55869177,
                0.08494262,
                0.25281882,
                8.9773321,
                1.77792146,
                0.10944497,
                0.22347975,
                0.07090735,
            ]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        logll = -870.3331059537576
        np.testing.assert_allclose(reg.logll, logll, RTOL)
        aic = 1756.6662119075154
        np.testing.assert_allclose(reg.aic, aic, RTOL)
        schwarz = 1783.481076975324
        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)
        chow_r = np.array(
            [
                [8.40437046, 0.0037432],
                [0.64080535, 0.42341932],
                [2.25389396, 0.13327865],
                [1.96544702, 0.16093197],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 25.367913028011799
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model2(self):
        reg = ML_Error_Regimes(
            self.y,
            self.x,
            self.regimes,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_w=self.w_name,
            name_ds=self.ds_name,
            name_regimes="CITCOU",
            regime_err_sep=True,
        )
        betas = np.array(
            [
                [3.66158216],
                [4.55700255],
                [-0.08045502],
                [0.44800318],
                [0.17774677],
                [33.3086368],
                [2.44709405],
                [-0.18803509],
                [0.68956598],
                [0.75599089],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array(
            [
                40.60994599,
                -7.25413138,
                -0.16605501,
                0.48961884,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([31.97771505])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([15.02228495])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([33.83065421])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        chow_r = np.array(
            [
                [6.88023639, 0.0087154],
                [0.90512612, 0.34141092],
                [0.75996258, 0.38334023],
                [0.56882946, 0.45072443],
                [12.18358581, 0.00048212],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 26.673798071789673
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)
        # Artficial:
        model = ML_Error_Regimes(
            self.y_a, self.x_a, self.regi_a, w=self.w_a, regime_err_sep=True
        )
        model1 = ML_Error(
            self.y_a[0 : (self.n2)].reshape((self.n2), 1),
            self.x_a[0 : (self.n2)],
            w=self.w_a1,
        )
        model2 = ML_Error(
            self.y_a[(self.n2) :].reshape((self.n2), 1),
            self.x_a[(self.n2) :],
            w=self.w_a1,
        )
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas, tbetas)
        vm = np.hstack((model1.vm.diagonal(), model2.vm.diagonal()))
        np.testing.assert_allclose(model.vm.diagonal(), vm, 4)


if __name__ == "__main__":
    unittest.main()



import unittest
import libpysal
from scipy import sparse
import numpy as np
from spreg.ml_lag_regimes import ML_Lag_Regimes
from libpysal.common import RTOL

from warnings import filterwarnings

filterwarnings("ignore", category=sparse.SparseEfficiencyWarning)


class TestMLError(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("baltim.dbf"), "r")
        self.ds_name = "baltim.dbf"
        self.y_name = "PRICE"
        self.y = np.array(db.by_col(self.y_name)).T
        self.y.shape = (len(self.y), 1)
        self.x_names = ["NROOM", "AGE", "SQFT"]
        self.x = np.array([db.by_col(var) for var in self.x_names]).T
        ww = libpysal.io.open(libpysal.examples.get_path("baltim_q.gal"))
        self.w = ww.read()
        ww.close()
        self.w_name = "baltim_q.gal"
        self.w.transform = "r"
        self.regimes = db.by_col("CITCOU")

    def test_model1(self):
        reg = ML_Lag_Regimes(
            self.y,
            self.x,
            self.regimes,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_w=self.w_name,
            name_ds=self.ds_name,
            name_regimes="CITCOU",
            regime_lag_sep=False,
        )
        betas = np.array(
            [
                [-15.00586577],
                [4.49600801],
                [-0.03180518],
                [0.34995882],
                [-4.54040395],
                [3.92187578],
                [-0.17021393],
                [0.81941371],
                [0.53850323],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([32.73718478])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([14.26281522])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 211
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 9
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([47.0])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[1.0, 4.0, 148.0, 11.25, 0.0, 0.0, 0.0, 0.0]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        e = np.array([29.45407124])
        np.testing.assert_allclose(reg.e_pred[0], e, RTOL)
        my = 44.307180094786695
        np.testing.assert_allclose(reg.mean_y, my)
        sy = 23.606076835380495
        np.testing.assert_allclose(reg.std_y, sy)
        vm = np.array(
            [
                47.42000914,
                2.39526578,
                0.00506895,
                0.06480022,
                69.67653371,
                3.20661492,
                0.01156766,
                0.04862014,
                0.00400775,
            ]
        )
        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL)
        sig2 = 200.04433357145007
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.6404460298085746
        np.testing.assert_allclose(reg.pr2, pr2)
        std_err = np.array(
            [
                6.88621878,
                1.54766462,
                0.07119654,
                0.25455888,
                8.34724707,
                1.79070235,
                0.10755305,
                0.22049975,
                0.0633068,
            ]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        logll = -864.98505596489736
        np.testing.assert_allclose(reg.logll, logll, RTOL)
        aic = 1747.9701119297947
        np.testing.assert_allclose(reg.aic, aic, RTOL)
        schwarz = 1778.1368351310794
        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)
        chow_r = np.array(
            [
                [1.00180776, 0.31687348],
                [0.05904944, 0.8080047],
                [1.16987812, 0.27942629],
                [1.95931177, 0.16158694],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 21.648337464039283
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model2(self):
        reg = ML_Lag_Regimes(
            self.y,
            self.x,
            self.regimes,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_w=self.w_name,
            name_ds=self.ds_name,
            name_regimes="CITCOU",
            regime_lag_sep=True,
        )
        betas = np.array(
            [
                [-0.71589799],
                [4.40910538],
                [-0.08652467],
                [0.46266265],
                [0.1627765],
                [-5.00594358],
                [2.91060349],
                [-0.18207394],
                [0.71129227],
                [0.66753263],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array(
            [
                55.3593679,
                -7.22927797,
                -0.19487326,
                0.6030953,
                -0.52249569,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([34.03630518])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([12.96369482])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([32.46466912])
        np.testing.assert_allclose(reg.e_pred[0], e, RTOL)
        chow_r = np.array(
            [
                [0.15654726, 0.69235548],
                [0.43533847, 0.509381],
                [0.60552514, 0.43647766],
                [0.59214981, 0.441589],
                [11.69437282, 0.00062689],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 21.978012275873063
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)


if __name__ == "__main__":
    unittest.main()



import unittest
import libpysal
import numpy as np
from spreg import error_sp_hom_regimes as SP
from spreg.error_sp_hom import GM_Error_Hom, GM_Endog_Error_Hom, GM_Combo_Hom
from libpysal.common import RTOL


class TestGM_Error_Hom_Regimes(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("CRIME"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("HOVAL"))
        X.append(db.by_col("INC"))
        self.X = np.array(X).T
        X2 = []
        X2.append(db.by_col("INC"))
        self.X2 = np.array(X2).T
        yd = []
        yd.append(db.by_col("HOVAL"))
        self.yd = np.array(yd).T
        q = []
        q.append(db.by_col("DISCBD"))
        self.q = np.array(q).T
        self.w = libpysal.weights.Queen.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"
        self.r_var = "NSA"
        self.regimes = db.by_col(self.r_var)
        # Artficial:
        n = 256
        self.n2 = n / 2
        self.x_a1 = np.random.uniform(-10, 10, (n, 1))
        self.x_a2 = np.random.uniform(1, 5, (n, 1))
        self.q_a = self.x_a2 + np.random.normal(0, 1, (n, 1))
        self.x_a = np.hstack((self.x_a1, self.x_a2))
        self.y_a = np.dot(
            np.hstack((np.ones((n, 1)), self.x_a)), np.array([[1], [0.5], [2]])
        ) + np.random.normal(0, 1, (n, 1))
        latt = int(np.sqrt(n))
        self.w_a = libpysal.weights.util.lat2W(latt, latt)
        self.w_a.transform = "r"
        self.regi_a = [0] * (n // 2) + [1] * (n // 2)  ##must be floos!
        self.w_a1 = libpysal.weights.util.lat2W(latt // 2, latt)
        self.w_a1.transform = "r"

    def test_model(self):
        reg = SP.GM_Error_Hom_Regimes(self.y, self.X, self.regimes, self.w, A1="het")
        betas = np.array(
            [
                [62.95986466],
                [-0.15660795],
                [-1.49054832],
                [60.98577615],
                [-0.3358993],
                [-0.82129289],
                [0.54033921],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([-2.19031456])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([17.91629456])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 6
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([15.72598])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[0.0, 0.0, 0.0, 1.0, 80.467003, 19.531]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        e = np.array([2.72131648])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        my = 35.128823897959187
        np.testing.assert_allclose(reg.mean_y, my, RTOL)
        sy = 16.732092091229699
        np.testing.assert_allclose(reg.std_y, sy, RTOL)
        vm = np.array(
            [
                49.16245801,
                -0.12493165,
                -1.89294614,
                5.71968257,
                -0.0571525,
                0.05745855,
                0.0,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        sig2 = 96.96108341267626
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.5515791216023577
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        std_err = np.array(
            [
                7.01159454,
                0.20701411,
                0.56905515,
                7.90537942,
                0.10268949,
                0.56660879,
                0.15659504,
            ]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        chow_r = np.array(
            [
                [0.03888544, 0.84367579],
                [0.61613446, 0.43248738],
                [0.72632441, 0.39407719],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 0.92133276766189676
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model_regi_error(self):
        # Artficial:
        model = SP.GM_Error_Hom_Regimes(
            self.y_a, self.x_a, self.regi_a, w=self.w_a, regime_err_sep=True, A1="het"
        )
        n2 = int(self.n2)
        model1 = GM_Error_Hom(
            self.y_a[0:n2].reshape(n2, 1), self.x_a[0:n2], w=self.w_a1, A1="het"
        )
        model2 = GM_Error_Hom(
            self.y_a[n2:].reshape(n2, 1), self.x_a[n2:], w=self.w_a1, A1="het"
        )
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas, tbetas, RTOL)
        vm = np.hstack((model1.vm.diagonal(), model2.vm.diagonal()))
        np.testing.assert_allclose(model.vm.diagonal(), vm, RTOL)
        # Columbus:
        reg = SP.GM_Error_Hom_Regimes(
            self.y, self.X2, self.regimes, self.w, regime_err_sep=True, A1="het"
        )
        betas = np.array(
            [
                [60.66668194],
                [-1.72708492],
                [0.62170311],
                [61.4526885],
                [-1.90700858],
                [0.1102755],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array([45.57956967, -1.65365774, 0.0, 0.0, 0.0, 0.0])
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([-8.48092392])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([24.20690392])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([-8.33982604])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        chow_r = np.array(
            [
                [0.0050892, 0.94312823],
                [0.05746619, 0.81054651],
                [1.65677138, 0.19803981],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 1.7914221673031792
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model_endog(self):
        reg = SP.GM_Endog_Error_Hom_Regimes(
            self.y, self.X2, self.yd, self.q, self.regimes, self.w, A1="het"
        )
        betas = np.array(
            [
                [77.26679984],
                [4.45992905],
                [78.59534391],
                [0.41432319],
                [-3.20196286],
                [-1.13672283],
                [0.22178164],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([20.50716917])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        e = np.array([25.22635318])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        predy = np.array([-4.78118917])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 6
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([15.72598])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[0.0, 0.0, 1.0, 19.531]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        yend = np.array([[0.0, 80.467003]])
        np.testing.assert_allclose(reg.yend[0].toarray(), yend, RTOL)
        z = np.array([[0.0, 0.0, 1.0, 19.531, 0.0, 80.467003]])
        np.testing.assert_allclose(reg.z[0].toarray(), z, RTOL)
        my = 35.128823897959187
        np.testing.assert_allclose(reg.mean_y, my, RTOL)
        sy = 16.732092091229699
        np.testing.assert_allclose(reg.std_y, sy, RTOL)
        vm = np.array(
            [
                403.76852704,
                69.06920553,
                19.8388512,
                3.62501395,
                -40.30472224,
                -1.6601927,
                -1.64319352,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        pr2 = 0.19776512679498906
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        sig2 = 644.23810259214
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        std_err = np.array(
            [
                20.09399231,
                7.03617703,
                23.64968032,
                2.176846,
                3.40352278,
                0.92377997,
                0.24462006,
            ]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        chow_r = np.array(
            [
                [0.00191145, 0.96512749],
                [0.31031517, 0.57748685],
                [0.34994619, 0.55414359],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 1.248410480025556
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model_endog_regi_error(self):
        # Columbus:
        reg = SP.GM_Endog_Error_Hom_Regimes(
            self.y,
            self.X2,
            self.yd,
            self.q,
            self.regimes,
            self.w,
            regime_err_sep=True,
            A1="het",
        )
        betas = np.array(
            [
                [7.92747424e01],
                [5.78086230e00],
                [-3.83173581e00],
                [2.14725610e-01],
                [8.26255251e01],
                [5.48294187e-01],
                [-1.28432891e00],
                [2.98658172e-02],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array(
            [867.50930457, 161.04430783, -92.35637083, -1.13838767, 0.0, 0.0, 0.0, 0.0]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([25.73781918])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([-10.01183918])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([26.41176711])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        chow_r = np.array(
            [
                [0.00909777, 0.92401124],
                [0.24034941, 0.62395386],
                [0.24322564, 0.62188603],
                [0.32572159, 0.5681893],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 1.4485058522307526
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)
        # Artficial:
        model = SP.GM_Endog_Error_Hom_Regimes(
            self.y_a,
            self.x_a1,
            yend=self.x_a2,
            q=self.q_a,
            regimes=self.regi_a,
            w=self.w_a,
            regime_err_sep=True,
            A1="het",
        )
        n2 = int(self.n2)
        model1 = GM_Endog_Error_Hom(
            self.y_a[0:n2].reshape(n2, 1),
            self.x_a1[0:n2],
            yend=self.x_a2[0:n2],
            q=self.q_a[0:n2],
            w=self.w_a1,
            A1="het",
        )
        model2 = GM_Endog_Error_Hom(
            self.y_a[n2:].reshape(n2, 1),
            self.x_a1[n2:],
            yend=self.x_a2[n2:],
            q=self.q_a[n2:],
            w=self.w_a1,
            A1="het",
        )
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas, tbetas)
        vm = np.hstack((model1.vm.diagonal(), model2.vm.diagonal()))
        np.testing.assert_allclose(model.vm.diagonal(), vm, RTOL)

    def test_model_combo(self):
        reg = SP.GM_Combo_Hom_Regimes(
            self.y, self.X2, self.regimes, self.yd, self.q, w=self.w, A1="het"
        )
        betas = np.array(
            [
                [36.93726782],
                [-0.829475],
                [30.86675168],
                [-0.72375344],
                [-0.30190094],
                [-0.22132895],
                [0.64190215],
                [-0.07314671],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([0.94039246])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        e_filtered = np.array([0.74211331])
        np.testing.assert_allclose(reg.e_filtered[0], e_filtered, RTOL)
        predy_e = np.array([18.68732105])
        np.testing.assert_allclose(reg.predy_e[0], predy_e, RTOL)
        predy = np.array([14.78558754])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 7
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([15.72598])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[0.0, 0.0, 1.0, 19.531]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        yend = np.array([[0.0, 80.467003, 24.7142675]])
        np.testing.assert_allclose(reg.yend[0].toarray(), yend, RTOL)
        z = np.array([[0.0, 0.0, 1.0, 19.531, 0.0, 80.467003, 24.7142675]])
        np.testing.assert_allclose(reg.z[0].toarray(), z, RTOL)
        my = 35.128823897959187
        np.testing.assert_allclose(reg.mean_y, my, RTOL)
        sy = 16.732092091229699
        np.testing.assert_allclose(reg.std_y, sy, RTOL)
        vm = np.array(
            [
                111.54419614,
                -0.23476709,
                83.37295278,
                -1.74452409,
                -1.60256796,
                -0.13151396,
                -1.43857915,
                2.19420848,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        sig2 = 95.57694234438294
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.6504148883591536
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        pr2_e = 0.5271368969923579
        np.testing.assert_allclose(reg.pr2_e, pr2_e, RTOL)
        std_err = np.array(
            [
                10.56144858,
                0.93986958,
                11.52977369,
                0.61000358,
                0.44419535,
                0.16191882,
                0.1630835,
                0.41107528,
            ]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        chow_r = np.array(
            [
                [0.47406771, 0.49112176],
                [0.00879838, 0.92526827],
                [0.02943577, 0.86377672],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 0.59098559257602923
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model_combo_regi_error(self):
        # Columbus:
        reg = SP.GM_Combo_Hom_Regimes(
            self.y,
            self.X2,
            self.regimes,
            self.yd,
            self.q,
            w=self.w,
            regime_lag_sep=True,
            regime_err_sep=True,
            A1="het",
        )
        betas = np.array(
            [
                [4.20115146e01],
                [-1.39171512e-01],
                [-6.53001838e-01],
                [5.47370644e-01],
                [2.61860326e-01],
                [3.42156975e01],
                [-1.52360889e-01],
                [-4.91752171e-01],
                [6.57331733e-01],
                [-2.68716241e-02],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array(
            [
                154.23356187,
                2.99104716,
                -3.29036767,
                -2.473113,
                1.65247551,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([7.81039418])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([7.91558582])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([7.60819283])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        chow_r = np.array(
            [
                [9.59590706e-02, 7.56733881e-01],
                [6.53130455e-05, 9.93551847e-01],
                [4.65270134e-02, 8.29220655e-01],
                [7.68939379e-02, 7.81551631e-01],
                [5.04560098e-01, 4.77503278e-01],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 0.74134991257940286
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)
        # Artficial:
        model = SP.GM_Combo_Hom_Regimes(
            self.y_a,
            self.x_a1,
            yend=self.x_a2,
            q=self.q_a,
            regimes=self.regi_a,
            w=self.w_a,
            regime_err_sep=True,
            regime_lag_sep=True,
            A1="het",
        )
        model1 = GM_Combo_Hom(
            self.y_a[0 : int(self.n2)].reshape(int(self.n2), 1),
            self.x_a1[0 : int(self.n2)],
            yend=self.x_a2[0 : int(self.n2)],
            q=self.q_a[0 : int(self.n2)],
            w=self.w_a1,
            A1="het",
        )
        model2 = GM_Combo_Hom(
            self.y_a[int(self.n2) :].reshape(int(self.n2), 1),
            self.x_a1[int(self.n2) :],
            yend=self.x_a2[int(self.n2) :],
            q=self.q_a[int(self.n2) :],
            w=self.w_a1,
            A1="het",
        )
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas, tbetas)
        vm = np.hstack((model1.vm.diagonal(), model2.vm.diagonal()))


if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import libpysal
from spreg.ols import OLS
from spreg.ols_regimes import OLS_Regimes
from libpysal.common import RTOL

PEGP = libpysal.examples.get_path


class TestOLS_regimes(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        self.y_var = "CRIME"
        self.y = np.array([db.by_col(self.y_var)]).reshape(49, 1)
        self.x_var = ["INC", "HOVAL"]
        self.x = np.array([db.by_col(name) for name in self.x_var]).T
        self.r_var = "NSA"
        self.regimes = db.by_col(self.r_var)
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_OLS(self):
        start_suppress = np.get_printoptions()["suppress"]
        np.set_printoptions(suppress=True)
        ols = OLS_Regimes(
            self.y,
            self.x,
            self.regimes,
            w=self.w,
            regime_err_sep=False,
            constant_regi="many",
            nonspat_diag=False,
            spat_diag=True,
            name_y=self.y_var,
            name_x=self.x_var,
            name_ds="columbus",
            name_regimes=self.r_var,
            name_w="columbus.gal",
        )
        # np.testing.assert_allclose(ols.aic, 408.73548964604873 ,RTOL)
        np.testing.assert_allclose(ols.ar2, 0.50761700679873101, RTOL)
        np.testing.assert_allclose(
            ols.betas,
            np.array(
                [
                    [68.78670869],
                    [-1.9864167],
                    [-0.10887962],
                    [67.73579559],
                    [-1.36937552],
                    [-0.31792362],
                ]
            ),
        )
        vm = np.array([48.81339213, -2.14959579, -0.19968157, 0.0, 0.0, 0.0])
        np.testing.assert_allclose(ols.vm[0], vm, RTOL)
        np.testing.assert_allclose(ols.lm_error, (5.92970357, 0.01488775), RTOL)
        np.testing.assert_allclose(ols.lm_lag, (8.78315751, 0.00304024), RTOL)
        np.testing.assert_allclose(ols.lm_sarma, (8.89955982, 0.01168114), RTOL)
        np.testing.assert_allclose(ols.mean_y, 35.1288238979591, RTOL)
        np.testing.assert_equal(ols.k, 6)
        np.testing.assert_equal(ols.kf, 0)
        np.testing.assert_equal(ols.kr, 3)
        np.testing.assert_equal(ols.n, 49)
        np.testing.assert_equal(ols.nr, 2)
        np.testing.assert_equal(ols.name_ds, "columbus")
        np.testing.assert_equal(ols.name_gwk, None)
        np.testing.assert_equal(ols.name_w, "columbus.gal")
        np.testing.assert_equal(
            ols.name_x,
            ["0_CONSTANT", "0_INC", "0_HOVAL", "1_CONSTANT", "1_INC", "1_HOVAL"],
        )
        np.testing.assert_equal(ols.name_y, "CRIME")
        np.testing.assert_allclose(ols.predy[3], np.array([51.05003696]), RTOL)
        np.testing.assert_allclose(ols.r2, 0.55890690192386316, RTOL)
        np.testing.assert_allclose(ols.rlm_error, (0.11640231, 0.73296972), RTOL)
        np.testing.assert_allclose(ols.rlm_lag, (2.96985625, 0.08482939), RTOL)
        np.testing.assert_equal(ols.robust, "unadjusted")
        np.testing.assert_allclose(ols.sig2, 137.84897351821013, RTOL)
        np.testing.assert_allclose(ols.sig2n, 120.96950737312316, RTOL)
        np.testing.assert_allclose(ols.t_stat[2][0], -0.43342216706091791, RTOL)
        np.testing.assert_allclose(ols.t_stat[2][1], 0.66687472578594531, RTOL)
        np.set_printoptions(suppress=start_suppress)

    """
    def test_OLS_regi(self):
        #Artficial:
        n = 256
        x1 = np.random.uniform(-10,10,(n,1))
        y = np.dot(np.hstack((np.ones((n,1)),x1)),np.array([[1],[0.5]])) + np.random.normal(0,1,(n,1))
        latt = int(np.sqrt(n))
        regi = [0]*(n/2) + [1]*(n/2)
        model = OLS_Regimes(y, x1, regimes=regi, regime_err_sep=True, sig2n_k=False)
        model1 = OLS(y[0:(n/2)].reshape((n/2),1), x1[0:(n/2)], sig2n_k=False)
        model2 = OLS(y[(n/2):n].reshape((n/2),1), x1[(n/2):n], sig2n_k=False)
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas,tbetas,RTOL)
        vm = np.hstack((model1.vm.diagonal(),model2.vm.diagonal()))
        np.testing.assert_allclose(model.vm.diagonal(), vm,RTOL)
        #Columbus:  
        reg = OLS_Regimes(self.y, self.x, self.regimes, w=self.w, constant_regi='many', nonspat_diag=True, spat_diag=True, name_y=self.y_var, name_x=self.x_var, name_ds='columbus', name_regimes=self.r_var, name_w='columbus.gal', regime_err_sep=True)        
        np.testing.assert_allclose(reg.multi[0].aic, 192.96044303402897 ,RTOL)
        tbetas = np.array([[ 68.78670869],
       [ -1.9864167 ],
       [ -0.10887962],
       [ 67.73579559],
       [ -1.36937552],
       [ -0.31792362]])
        np.testing.assert_allclose(tbetas, reg.betas,RTOL)
        vm = np.array([ 41.68828023,  -1.83582717,  -0.17053478,   0.        ,
         0.        ,   0.        ])
        np.testing.assert_allclose(reg.vm[0], vm,RTOL)
        u_3 = np.array([[ 0.31781838],
       [-5.6905584 ],
       [-6.8819715 ]])
        np.testing.assert_allclose(reg.u[0:3], u_3,RTOL)
        predy_3 = np.array([[ 15.40816162],
       [ 24.4923124 ],
       [ 37.5087525 ]])
        np.testing.assert_allclose(reg.predy[0:3], predy_3,RTOL)
        chow_regi = np.array([[ 0.01002733,  0.92023592],
       [ 0.46017009,  0.49754449],
       [ 0.60732697,  0.43579603]])
        np.testing.assert_allclose(reg.chow.regi, chow_regi,RTOL)
        np.testing.assert_allclose(reg.chow.joint[0], 0.67787986791767096,RTOL)
    """

    def test_OLS_fixed(self):
        start_suppress = np.get_printoptions()["suppress"]
        np.set_printoptions(suppress=True)
        ols = OLS_Regimes(
            self.y,
            self.x,
            self.regimes,
            w=self.w,
            cols2regi=[False, True],
            regime_err_sep=True,
            constant_regi="one",
            nonspat_diag=False,
            spat_diag=True,
            name_y=self.y_var,
            name_x=self.x_var,
            name_ds="columbus",
            name_regimes=self.r_var,
            name_w="columbus.gal",
        )
        np.testing.assert_allclose(
            ols.betas,
            np.array([[-0.24385565], [-0.26335026], [68.89701137], [-1.67389685]]),
            RTOL,
        )
        vm = np.array([0.02354271, 0.01246677, 0.00424658, -0.04921356])
        np.testing.assert_allclose(ols.vm[0], vm, RTOL)
        np.testing.assert_allclose(ols.lm_error, (5.62668744, 0.01768903), RTOL)
        np.testing.assert_allclose(ols.lm_lag, (9.43264957, 0.00213156), RTOL)
        np.testing.assert_allclose(ols.mean_y, 35.12882389795919, RTOL)
        np.testing.assert_equal(ols.kf, 2)
        np.testing.assert_equal(ols.kr, 1)
        np.testing.assert_equal(ols.n, 49)
        np.testing.assert_equal(ols.nr, 2)
        np.testing.assert_equal(ols.name_ds, "columbus")
        np.testing.assert_equal(ols.name_gwk, None)
        np.testing.assert_equal(ols.name_w, "columbus.gal")
        np.testing.assert_equal(
            ols.name_x, ["0_HOVAL", "1_HOVAL", "_Global_CONSTANT", "_Global_INC"]
        )
        np.testing.assert_equal(ols.name_y, "CRIME")
        np.testing.assert_allclose(ols.predy[3], np.array([52.65974636]), RTOL)
        np.testing.assert_allclose(ols.r2, 0.5525561183786056, RTOL)
        np.testing.assert_equal(ols.robust, "unadjusted")
        np.testing.assert_allclose(ols.t_stat[2][0], 13.848705206463748, RTOL)
        np.testing.assert_allclose(ols.t_stat[2][1], 7.776650625274256e-18, RTOL)
        np.set_printoptions(suppress=start_suppress)


if __name__ == "__main__":
    unittest.main()



import unittest
import libpysal
import numpy as np
from spreg.panel_fe import Panel_FE_Lag, Panel_FE_Error
from libpysal.common import RTOL


class Test_Panel_FE_Lag(unittest.TestCase):
    def setUp(self):
        self.ds_name = "NCOVR"
        nat = libpysal.examples.load_example(self.ds_name)
        self.db = libpysal.io.open(nat.get_path("NAT.dbf"), "r")
        nat_shp = libpysal.examples.get_path("NAT.shp")
        self.w = libpysal.weights.Queen.from_shapefile(nat_shp)
        self.w.transform = "r"
        self.y_name = ["HR70", "HR80", "HR90"]
        self.x_names = ["RD70", "RD80", "RD90", "PS70", "PS80", "PS90"]
        self.y = np.array([self.db.by_col(name) for name in self.y_name]).T
        self.x = np.array([self.db.by_col(name) for name in self.x_names]).T

    def test_Panel(self):
        reg = Panel_FE_Lag(
            self.y,
            self.x,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_ds=self.ds_name,
        )
        betas = np.array([[0.80058859], [-2.60035236], [0.19030424]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([-2.70317346])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([-0.24876891])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        vm = np.array([0.02606527, 0.24359025, 0.00025597])
        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL)
        sig2 = np.array([[14.93535335]])
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.03191868031797557
        np.testing.assert_allclose(reg.pr2, pr2)
        std_err = np.array([0.16144743, 0.49354863, 0.01599908])
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        logll = -67936.53303509377
        np.testing.assert_allclose(reg.logll, logll, RTOL)
        aic = 135879.06607018755
        np.testing.assert_allclose(reg.aic, aic, RTOL)
        schwarz = 135900.46482786257
        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)


class Test_Panel_FE_Error(unittest.TestCase):
    def setUp(self):
        self.ds_name = "NCOVR"
        nat = libpysal.examples.load_example(self.ds_name)
        self.db = libpysal.io.open(nat.get_path("NAT.dbf"), "r")
        nat_shp = libpysal.examples.get_path("NAT.shp")
        self.w = libpysal.weights.Queen.from_shapefile(nat_shp)
        self.w.transform = "r"
        self.y_name = ["HR70", "HR80", "HR90"]
        self.x_names = ["RD70", "RD80", "RD90", "PS70", "PS80", "PS90"]
        self.y = np.array([self.db.by_col(name) for name in self.y_name]).T
        self.x = np.array([self.db.by_col(name) for name in self.x_names]).T

    def test_Panel(self):
        reg = Panel_FE_Error(
            self.y,
            self.x,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_ds=self.ds_name,
        )
        betas = np.array([[0.86979232], [-2.96606744], [0.19434604]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([-3.02217669])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([0.07023431])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        vm = np.array([0.02951625, 0.29645666, 0.00025681])
        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL)
        sig2 = np.array([[14.92289858]])
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.008359611052787335
        np.testing.assert_allclose(reg.pr2, pr2)
        std_err = np.array([0.17180294, 0.54447834, 0.01602534])
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        logll = -67934.00512805565
        np.testing.assert_allclose(reg.logll, logll, RTOL)
        aic = 135872.0102561113
        np.testing.assert_allclose(reg.aic, aic, RTOL)
        schwarz = 135886.27609456133
        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)


if __name__ == "__main__":
    unittest.main()



import unittest
import libpysal
import numpy as np
import pandas as pd
from spreg.diagnostics_panel import panel_LMlag, panel_LMerror, panel_rLMlag
from spreg.diagnostics_panel import panel_rLMerror, panel_Hausman
from spreg.panel_fe import Panel_FE_Lag, Panel_FE_Error
from spreg.panel_re import Panel_RE_Lag, Panel_RE_Error
from libpysal.common import RTOL
from libpysal.weights import w_subset


class Test_Panel_Diagnostics(unittest.TestCase):
    def setUp(self):
        self.ds_name = "NCOVR"
        nat = libpysal.examples.load_example(self.ds_name)
        self.db = libpysal.io.open(nat.get_path("NAT.dbf"), "r")
        nat_shp = libpysal.examples.get_path("NAT.shp")
        w_full = libpysal.weights.Queen.from_shapefile(nat_shp)
        self.y_name = ["HR70", "HR80", "HR90"]
        self.x_names = ["RD70", "RD80", "RD90", "PS70", "PS80", "PS90"]
        c_names = ["STATE_NAME", "FIPSNO"]
        y_full = [self.db.by_col(name) for name in self.y_name]
        y_full = np.array(y_full).T
        x_full = [self.db.by_col(name) for name in self.x_names]
        x_full = np.array(x_full).T
        c_full = [self.db.by_col(name) for name in c_names]
        c_full = pd.DataFrame(c_full, index=c_names).T
        filter_states = ["Kansas", "Missouri", "Oklahoma", "Arkansas"]
        filter_counties = c_full[c_full["STATE_NAME"].isin(filter_states)]
        filter_counties = filter_counties["FIPSNO"].values
        counties = np.array(self.db.by_col("FIPSNO"))
        subid = np.where(np.isin(counties, filter_counties))[0]
        self.w = w_subset(w_full, subid)
        self.w.transform = "r"
        self.y = y_full[
            subid,
        ]
        self.x = x_full[
            subid,
        ]

    def test_LM(self):
        lmlag = panel_LMlag(self.y, self.x, self.w)
        exp = np.array([1.472807526666869, 0.22490325114767176])
        np.testing.assert_allclose(lmlag, exp, RTOL)
        lmerror = panel_LMerror(self.y, self.x, self.w)
        exp = np.array([81.69630396101608, 1.5868998506678388e-19])
        np.testing.assert_allclose(lmerror, exp, RTOL)
        rlmlag = panel_rLMlag(self.y, self.x, self.w)
        exp = np.array([2.5125780962741793, 0.11294102977710921])
        np.testing.assert_allclose(rlmlag, exp, RTOL)
        rlmerror = panel_rLMerror(self.y, self.x, self.w)
        exp = np.array([32.14155241279442, 1.4333858484607395e-08])
        np.testing.assert_allclose(rlmerror, exp, RTOL)

    def test_Hausman(self):
        fe_lag = Panel_FE_Lag(self.y, self.x, self.w)
#        fe_error = Panel_FE_Error(self.y, self.x, self.w)
        re_lag = Panel_RE_Lag(self.y, self.x, self.w)
#        re_error = Panel_RE_Error(self.y, self.x, self.w)
        Hlag = panel_Hausman(fe_lag, re_lag)
        exp = np.array([-67.26822586935438, 1.0])
        np.testing.assert_allclose(Hlag, exp, RTOL)
#        Herror = panel_Hausman(fe_error, re_error)
#        exp = np.array([-84.38351088621853, 1.0])
#        np.testing.assert_allclose(Herror, exp, RTOL)


if __name__ == "__main__":
    unittest.main()



import unittest
import libpysal
import numpy as np
from spreg import error_sp_het_regimes as SP
from spreg.error_sp_het import GM_Error_Het, GM_Endog_Error_Het, GM_Combo_Het
from libpysal.common import RTOL


class TestGM_Error_Het_Regimes(unittest.TestCase):
    def setUp(self):
        # Columbus:
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("CRIME"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("HOVAL"))
        X.append(db.by_col("INC"))
        self.X = np.array(X).T
        X2 = []
        X2.append(db.by_col("INC"))
        self.X2 = np.array(X2).T
        yd = []
        yd.append(db.by_col("HOVAL"))
        self.yd = np.array(yd).T
        q = []
        q.append(db.by_col("DISCBD"))
        self.q = np.array(q).T
        self.w = libpysal.weights.Queen.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"
        self.r_var = "NSA"
        self.regimes = db.by_col(self.r_var)
        # Artficial:
        n = 256
        self.n2 = n / 2
        self.x_a1 = np.random.uniform(-10, 10, (n, 1))
        self.x_a2 = np.random.uniform(1, 5, (n, 1))
        self.q_a = self.x_a2 + np.random.normal(0, 1, (n, 1))
        self.x_a = np.hstack((self.x_a1, self.x_a2))
        self.y_a = np.dot(
            np.hstack((np.ones((n, 1)), self.x_a)), np.array([[1], [0.5], [2]])
        ) + np.random.normal(0, 1, (n, 1))
        latt = int(np.sqrt(n))
        self.w_a = libpysal.weights.util.lat2W(latt, latt)
        self.w_a.transform = "r"
        self.regi_a = [0] * (n // 2) + [1] * (n // 2)
        self.w_a1 = libpysal.weights.util.lat2W(latt // 2, latt)
        self.w_a1.transform = "r"

    def test_model(self):
        reg = SP.GM_Error_Het_Regimes(self.y, self.X, self.regimes, self.w)
        betas = np.array(
            [
                [62.95986466],
                [-0.15660795],
                [-1.49054832],
                [60.98577615],
                [-0.3358993],
                [-0.82129289],
                [0.54662719],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([-2.19031456])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([17.91629456])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 6
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([15.72598])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[0.0, 0.0, 0.0, 1.0, 80.467003, 19.531]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        e = np.array([2.77847355])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        my = 35.128823897959187
        np.testing.assert_allclose(reg.mean_y, my, RTOL)
        sy = 16.732092091229699
        np.testing.assert_allclose(reg.std_y, sy, RTOL)
        vm = np.array(
            [
                3.86154100e01,
                -2.51553730e-01,
                -8.20138673e-01,
                1.71714184e00,
                -1.94929113e-02,
                1.23118051e-01,
                0.00000000e00,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        pr2 = 0.5515791216043385
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        std_err = np.array(
            [
                6.21412987,
                0.15340022,
                0.44060473,
                7.6032169,
                0.19353719,
                0.73621596,
                0.13968272,
            ]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        chow_r = np.array(
            [
                [0.04190799, 0.83779526],
                [0.5736724, 0.44880328],
                [0.62498575, 0.42920056],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 0.72341901308525713
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model_regi_error(self):
        # Columbus:
        reg = SP.GM_Error_Het_Regimes(
            self.y, self.X, self.regimes, self.w, regime_err_sep=True
        )
        betas = np.array(
            [
                [60.74090229],
                [-0.17492294],
                [-1.33383387],
                [0.68303064],
                [66.30374279],
                [-0.31841139],
                [-1.27502813],
                [0.11515312],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array([44.9411672, -0.34343354, -0.39946055, 0.0, 0.0, 0.0, 0.0, 0.0])
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([-0.05357818])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([15.77955818])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([0.70542044])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        chow_r = np.array(
            [
                [3.11061225e-01, 5.77029704e-01],
                [3.39747489e-01, 5.59975012e-01],
                [3.86371771e-03, 9.50436364e-01],
                [4.02884201e00, 4.47286322e-02],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 4.7467070503995412
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)
        # Artficial:
        model = SP.GM_Error_Het_Regimes(
            self.y_a, self.x_a, self.regi_a, w=self.w_a, regime_err_sep=True
        )
        model1 = GM_Error_Het(
            self.y_a[0 : int(self.n2)].reshape(int(self.n2), 1),
            self.x_a[0 : int(self.n2)],
            w=self.w_a1,
        )
        model2 = GM_Error_Het(
            self.y_a[int(self.n2) :].reshape(int(self.n2), 1),
            self.x_a[int(self.n2) :],
            w=self.w_a1,
        )
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas, tbetas, RTOL)
        vm = np.hstack((model1.vm.diagonal(), model2.vm.diagonal()))
        np.testing.assert_allclose(model.vm.diagonal(), vm, RTOL)

    def test_model_endog(self):
        reg = SP.GM_Endog_Error_Het_Regimes(
            self.y, self.X2, self.yd, self.q, self.regimes, self.w
        )
        betas = np.array(
            [
                [77.26679984],
                [4.45992905],
                [78.59534391],
                [0.41432319],
                [-3.20196286],
                [-1.13672283],
                [0.2174965],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([20.50716917])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        e = np.array([25.13517175])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        predy = np.array([-4.78118917])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 6
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([15.72598])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[0.0, 0.0, 1.0, 19.531]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        yend = np.array([[0.0, 80.467003]])
        np.testing.assert_allclose(reg.yend[0].toarray(), yend, RTOL)
        z = np.array([[0.0, 0.0, 1.0, 19.531, 0.0, 80.467003]])
        np.testing.assert_allclose(reg.z[0].toarray(), z, RTOL)
        my = 35.128823897959187
        np.testing.assert_allclose(reg.mean_y, my, RTOL)
        sy = 16.732092091229699
        np.testing.assert_allclose(reg.std_y, sy, RTOL)
        vm = np.array(
            [
                509.66122149,
                150.5845341,
                9.64413821,
                5.54782831,
                -80.95846045,
                -2.25308524,
                -3.2045214,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        pr2 = 0.19776512679331681
        np.testing.assert_allclose(reg.pr2, pr2)
        std_err = np.array(
            [
                22.57567765,
                11.34616946,
                17.43881791,
                1.30953812,
                5.4830829,
                0.74634612,
                0.29973079,
            ]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        chow_r = np.array(
            [[0.0022216, 0.96240654], [0.13127347, 0.7171153], [0.14367307, 0.70465645]]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 1.2329971019087163
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model_endog_regi_error(self):
        # Columbus:
        reg = SP.GM_Endog_Error_Het_Regimes(
            self.y, self.X2, self.yd, self.q, self.regimes, self.w, regime_err_sep=True
        )
        betas = np.array(
            [
                [7.92747424e01],
                [5.78086230e00],
                [-3.83173581e00],
                [2.23210962e-01],
                [8.26255251e01],
                [5.48294187e-01],
                [-1.28432891e00],
                [3.57661629e-02],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array(
            [
                7.55988579e02,
                2.53659722e02,
                -1.34288316e02,
                -2.66141766e-01,
                0.00000000e00,
                0.00000000e00,
                0.00000000e00,
                0.00000000e00,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([25.73781918])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([-10.01183918])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([26.5449135])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        chow_r = np.array(
            [
                [0.00998573, 0.92040097],
                [0.12660165, 0.72198192],
                [0.12737281, 0.72117171],
                [0.43507956, 0.50950696],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 1.3756768204399892
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)
        # Artficial:
        model = SP.GM_Endog_Error_Het_Regimes(
            self.y_a,
            self.x_a1,
            yend=self.x_a2,
            q=self.q_a,
            regimes=self.regi_a,
            w=self.w_a,
            regime_err_sep=True,
        )
        model1 = GM_Endog_Error_Het(
            self.y_a[0 : int(self.n2)].reshape(int(self.n2), 1),
            self.x_a1[0 : int(self.n2)],
            yend=self.x_a2[0 : int(self.n2)],
            q=self.q_a[0 : int(self.n2)],
            w=self.w_a1,
        )
        model2 = GM_Endog_Error_Het(
            self.y_a[int(self.n2) :].reshape(int(self.n2), 1),
            self.x_a1[int(self.n2) :],
            yend=self.x_a2[int(self.n2) :],
            q=self.q_a[int(self.n2) :],
            w=self.w_a1,
        )
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas, tbetas, RTOL)
        vm = np.hstack((model1.vm.diagonal(), model2.vm.diagonal()))
        np.testing.assert_allclose(model.vm.diagonal(), vm, RTOL)

    def test_model_combo(self):
        reg = SP.GM_Combo_Het_Regimes(
            self.y, self.X2, self.regimes, self.yd, self.q, w=self.w
        )
        betas = np.array(
            [
                [3.69372678e01],
                [-8.29474998e-01],
                [3.08667517e01],
                [-7.23753444e-01],
                [-3.01900940e-01],
                [-2.21328949e-01],
                [6.41902155e-01],
                [-2.45714919e-02],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([0.94039246])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        e_filtered = np.array([0.8737864])
        np.testing.assert_allclose(reg.e_filtered[0], e_filtered, RTOL)
        predy_e = np.array([18.68732105])
        np.testing.assert_allclose(reg.predy_e[0], predy_e, RTOL)
        predy = np.array([14.78558754])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 49
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 7
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([15.72598])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([[0.0, 0.0, 1.0, 19.531]])
        np.testing.assert_allclose(reg.x[0].toarray(), x, RTOL)
        yend = np.array([[0.0, 80.467003, 24.7142675]])
        np.testing.assert_allclose(reg.yend[0].toarray(), yend, RTOL)
        z = np.array([[0.0, 0.0, 1.0, 19.531, 0.0, 80.467003, 24.7142675]])
        np.testing.assert_allclose(reg.z[0].toarray(), z, RTOL)
        my = 35.128823897959187
        np.testing.assert_allclose(reg.mean_y, my, RTOL)
        sy = 16.732092091229699
        np.testing.assert_allclose(reg.std_y, sy, RTOL)
        vm = np.array(
            [
                71.26851365,
                -0.58278032,
                50.53169815,
                -0.74561147,
                -0.79510274,
                -0.10823496,
                -0.98141395,
                1.16575965,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        pr2 = 0.6504148883602958
        np.testing.assert_allclose(reg.pr2, pr2, RTOL)
        pr2_e = 0.527136896994038
        np.testing.assert_allclose(reg.pr2_e, pr2_e, RTOL)
        std_err = np.array(
            [
                8.44206809,
                0.72363219,
                9.85790968,
                0.77218082,
                0.34084146,
                0.21752916,
                0.14371614,
                0.39226478,
            ]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        chow_r = np.array(
            [
                [0.54688708, 0.45959243],
                [0.01035136, 0.91896175],
                [0.03981108, 0.84185042],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 0.78070369988354349
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)

    def test_model_combo_regi_error(self):
        # Columbus:
        reg = SP.GM_Combo_Het_Regimes(
            self.y,
            self.X2,
            self.regimes,
            self.yd,
            self.q,
            w=self.w,
            regime_lag_sep=True,
            regime_err_sep=True,
        )
        betas = np.array(
            [
                [42.01151458],
                [-0.13917151],
                [-0.65300184],
                [0.54737064],
                [0.2629229],
                [34.21569751],
                [-0.15236089],
                [-0.49175217],
                [0.65733173],
                [-0.07713581],
            ]
        )
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        vm = np.array(
            [
                77.49519689,
                0.57226879,
                -1.18856422,
                -1.28088712,
                0.866752,
                0.0,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        np.testing.assert_allclose(reg.vm[0], vm, RTOL)
        u = np.array([7.81039418])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([7.91558582])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        e = np.array([7.22996911])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        chow_r = np.array(
            [
                [1.90869079e-01, 6.62194273e-01],
                [4.56118982e-05, 9.94611401e-01],
                [3.12104263e-02, 8.59771748e-01],
                [1.56368204e-01, 6.92522476e-01],
                [7.52928732e-01, 3.85550558e-01],
            ]
        )
        np.testing.assert_allclose(reg.chow.regi, chow_r, RTOL)
        chow_j = 1.1316136604755913
        np.testing.assert_allclose(reg.chow.joint[0], chow_j, RTOL)
        # Artficial:
        model = SP.GM_Combo_Het_Regimes(
            self.y_a,
            self.x_a1,
            yend=self.x_a2,
            q=self.q_a,
            regimes=self.regi_a,
            w=self.w_a,
            regime_err_sep=True,
            regime_lag_sep=True,
        )
        model1 = GM_Combo_Het(
            self.y_a[0 : int(self.n2)].reshape(int(self.n2), 1),
            self.x_a1[0 : int(self.n2)],
            yend=self.x_a2[0 : int(self.n2)],
            q=self.q_a[0 : int(self.n2)],
            w=self.w_a1,
        )
        model2 = GM_Combo_Het(
            self.y_a[int(self.n2) :].reshape(int(self.n2), 1),
            self.x_a1[int(self.n2) :],
            yend=self.x_a2[int(self.n2) :],
            q=self.q_a[int(self.n2) :],
            w=self.w_a1,
        )
        tbetas = np.vstack((model1.betas, model2.betas))
        np.testing.assert_allclose(model.betas, tbetas, RTOL)
        vm = np.hstack((model1.vm.diagonal(), model2.vm.diagonal()))
        # was this supposed to get tested?


if __name__ == "__main__":
    unittest.main()



import unittest
import libpysal
from libpysal.examples import load_example
from libpysal.weights import Queen
import numpy as np
from scipy import sparse
from spreg.ml_error import ML_Error
from libpysal.common import RTOL, ATOL
from warnings import filterwarnings

filterwarnings("ignore", category=sparse.SparseEfficiencyWarning)
filterwarnings("ignore", message="^Method 'bounded' does not support")


class TestMLError(unittest.TestCase):
    def setUp(self):
        south = load_example("South")
        db = libpysal.io.open(south.get_path("south.dbf"), "r")
        self.y_name = "HR90"
        self.y = np.array(db.by_col(self.y_name))
        self.y.shape = (len(self.y), 1)
        self.x_names = ["RD90", "PS90", "UE90", "DV90"]
        self.x = np.array([db.by_col(var) for var in self.x_names]).T
        self.w = Queen.from_shapefile(south.get_path("south.shp"))
        self.w.transform = "r"

    def _estimate_and_compare(self, method="FULL", RTOL=RTOL):
        reg = ML_Error(
            self.y,
            self.x,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_w="south_q.gal",
            method=method,
        )
        betas = np.array([[6.1492], [4.4024], [1.7784], [-0.3781], [0.4858], [0.2991]])
        np.testing.assert_allclose(reg.betas, betas, RTOL + 0.0001)
        u = np.array([-5.97649777])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([6.92258051])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 1412
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 5
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([0.94608274])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, -0.39902838, 0.89645344, 6.85780705, 7.2636377])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        e = np.array([-4.92843327])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        my = 9.5492931620846928
        np.testing.assert_allclose(reg.mean_y, my)
        sy = 7.0388508798387219
        np.testing.assert_allclose(reg.std_y, sy)
        vm = np.array(
            [1.06476526, 0.05548248, 0.04544514, 0.00614425, 0.01481356, 0.00143001]
        )
        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL)
        sig2 = np.array([[32.40685441]])
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.3057664820364818
        np.testing.assert_allclose(reg.pr2, pr2)
        std_err = np.array(
            [1.03187463, 0.23554719, 0.21317867, 0.07838525, 0.12171098, 0.03781546]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        z_stat = [
            (5.9592751097983534, 2.5335926307459251e-09),
            (18.690182928021841, 5.9508619446611137e-78),
            (8.3421632936950338, 7.2943630281051907e-17),
            (-4.8232686291115678, 1.4122456582517099e-06),
            (3.9913060809142995, 6.5710406838016854e-05),
            (7.9088780724028922, 2.5971882547279339e-15),
        ]
        np.testing.assert_allclose(reg.z_stat, z_stat, RTOL, atol=ATOL)
        logll = -4471.407066887894
        np.testing.assert_allclose(reg.logll, logll, RTOL)
        aic = 8952.8141337757879
        np.testing.assert_allclose(reg.aic, aic, RTOL)
        schwarz = 8979.0779458660545
        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)

    def test_dense(self):
        self._estimate_and_compare(method="FULL")

    def test_LU(self):
        self._estimate_and_compare(method="LU", RTOL=RTOL * 10)

    def test_ord(self):
        reg = ML_Error(
            self.y,
            self.x,
            w=self.w,
            name_y=self.y_name,
            name_x=self.x_names,
            name_w="south_q.gal",
            method="ORD",
        )
        betas = np.array([[6.1492], [4.4024], [1.7784], [-0.3781], [0.4858], [0.2991]])
        np.testing.assert_allclose(reg.betas, betas, RTOL + 0.0001)
        u = np.array([-5.97649777])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        predy = np.array([6.92258051])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        n = 1412
        np.testing.assert_allclose(reg.n, n, RTOL)
        k = 5
        np.testing.assert_allclose(reg.k, k, RTOL)
        y = np.array([0.94608274])
        np.testing.assert_allclose(reg.y[0], y, RTOL)
        x = np.array([1.0, -0.39902838, 0.89645344, 6.85780705, 7.2636377])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        e = np.array([-4.92843327])
        np.testing.assert_allclose(reg.e_filtered[0], e, RTOL)
        my = 9.5492931620846928
        np.testing.assert_allclose(reg.mean_y, my)
        sy = 7.0388508798387219
        np.testing.assert_allclose(reg.std_y, sy)
        vm = np.array(
            [1.06476526, 0.05548248, 0.04544514, 0.00614425, 0.01481356, 0.001501]
        )
        np.testing.assert_allclose(reg.vm.diagonal(), vm, RTOL * 10)
        sig2 = np.array([[32.40685441]])
        np.testing.assert_allclose(reg.sig2, sig2, RTOL)
        pr2 = 0.3057664820364818
        np.testing.assert_allclose(reg.pr2, pr2)
        std_err = np.array(
            [1.03187463, 0.23554719, 0.21317867, 0.07838525, 0.12171098, 0.038744]
        )
        np.testing.assert_allclose(reg.std_err, std_err, RTOL * 10)
        z_stat = [
            (5.95927510, 2.5335927e-09),
            (18.6901829, 5.9508630e-78),
            (8.34216329, 7.2943634e-17),
            (-4.8232686, 1.4122457e-06),
            (3.99130608, 6.5710407e-05),
            (7.71923784, 1.1702739e-14),
        ]
        np.testing.assert_allclose(reg.z_stat, z_stat, rtol=RTOL, atol=ATOL)
        logll = -4471.407066887894
        np.testing.assert_allclose(reg.logll, logll, RTOL)
        aic = 8952.8141337757879
        np.testing.assert_allclose(reg.aic, aic, RTOL)
        schwarz = 8979.0779458660545
        np.testing.assert_allclose(reg.schwarz, schwarz, RTOL)


if __name__ == "__main__":
    unittest.main()



"""
Unittests for spreg.error_sp_hom module

"""
import unittest
import libpysal
from spreg import error_sp_hom as HOM
import numpy as np
from libpysal.common import RTOL
import spreg


class BaseGM_Error_Hom_Tester(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        X.append(db.by_col("CRIME"))
        self.X = np.array(X).T
        self.X = np.hstack((np.ones(self.y.shape), self.X))
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = HOM.BaseGM_Error_Hom(self.y, self.X, self.w.sparse, A1="hom_sc")
        np.testing.assert_allclose(reg.y[0], np.array([80.467003]), RTOL)
        x = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        betas = np.array([[47.9478524], [0.70633223], [-0.55595633], [0.41288558]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        np.testing.assert_allclose(reg.u[0], np.array([27.466734]), RTOL)
        np.testing.assert_allclose(reg.e_filtered[0], np.array([32.37298547]), RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        np.testing.assert_allclose(reg.predy[0], np.array([53.000269]), RTOL)
        np.testing.assert_allclose(reg.n, 49, RTOL)
        np.testing.assert_allclose(reg.k, 3, RTOL)
        sig2 = 189.94459439729718
        np.testing.assert_allclose(reg.sig2, sig2)
        vm = np.array(
            [
                [1.51340717e02, -5.29057506e00, -1.85654540e00, -2.39139054e-03],
                [-5.29057506e00, 2.46669610e-01, 5.14259101e-02, 3.19241302e-04],
                [-1.85654540e00, 5.14259101e-02, 3.20510550e-02, -5.95640240e-05],
                [-2.39139054e-03, 3.19241302e-04, -5.95640240e-05, 3.36690159e-02],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        xtx = np.array(
            [
                [4.90000000e01, 7.04371999e02, 1.72131237e03],
                [7.04371999e02, 1.16866734e04, 2.15575320e04],
                [1.72131237e03, 2.15575320e04, 7.39058986e04],
            ]
        )
        np.testing.assert_allclose(reg.xtx, xtx, RTOL)


class GM_Error_Hom_Tester(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        X.append(db.by_col("CRIME"))
        self.X = np.array(X).T
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = HOM.GM_Error_Hom(self.y, self.X, self.w, A1="hom_sc")
        np.testing.assert_allclose(reg.y[0], np.array([80.467003]), RTOL)
        x = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        betas = np.array([[47.9478524], [0.70633223], [-0.55595633], [0.41288558]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        np.testing.assert_allclose(reg.u[0], np.array([27.46673388]), RTOL)
        np.testing.assert_allclose(reg.e_filtered[0], np.array([32.37298547]), RTOL)
        np.testing.assert_allclose(reg.predy[0], np.array([53.00026912]), RTOL)
        np.testing.assert_allclose(reg.n, 49, RTOL)
        np.testing.assert_allclose(reg.k, 3, RTOL)
        vm = np.array(
            [
                [1.51340717e02, -5.29057506e00, -1.85654540e00, -2.39139054e-03],
                [-5.29057506e00, 2.46669610e-01, 5.14259101e-02, 3.19241302e-04],
                [-1.85654540e00, 5.14259101e-02, 3.20510550e-02, -5.95640240e-05],
                [-2.39139054e-03, 3.19241302e-04, -5.95640240e-05, 3.36690159e-02],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        np.testing.assert_allclose(reg.iteration, 1, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        std_y = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, std_y)
        pr2 = 0.34950977055969729
        np.testing.assert_allclose(reg.pr2, pr2)
        sig2 = 189.94459439729718
        np.testing.assert_allclose(reg.sig2, sig2)
        std_err = np.array([12.30206149, 0.49665844, 0.17902808, 0.18349119])
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        z_stat = np.array(
            [
                [3.89754616e00, 9.71723059e-05],
                [1.42216900e00, 1.54977196e-01],
                [-3.10541409e00, 1.90012806e-03],
                [2.25016500e00, 2.44384731e-02],
            ]
        )
        np.testing.assert_allclose(reg.z_stat, z_stat, RTOL)
        xtx = np.array(
            [
                [4.90000000e01, 7.04371999e02, 1.72131237e03],
                [7.04371999e02, 1.16866734e04, 2.15575320e04],
                [1.72131237e03, 2.15575320e04, 7.39058986e04],
            ]
        )
        np.testing.assert_allclose(reg.xtx, xtx, RTOL)


class BaseGM_Endog_Error_Hom_Tester(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        self.X = np.array(X).T
        self.X = np.hstack((np.ones(self.y.shape), self.X))
        yd = []
        yd.append(db.by_col("CRIME"))
        self.yd = np.array(yd).T
        q = []
        q.append(db.by_col("DISCBD"))
        self.q = np.array(q).T
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = HOM.BaseGM_Endog_Error_Hom(
            self.y, self.X, self.yd, self.q, self.w.sparse, A1="hom_sc"
        )
        np.testing.assert_allclose(reg.y[0], np.array([80.467003]), RTOL)
        x = np.array([1.0, 19.531])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        z = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.z[0], z, RTOL)
        h = np.array([1.0, 19.531, 5.03])
        np.testing.assert_allclose(reg.h[0], h, RTOL)
        yend = np.array([15.72598])
        np.testing.assert_allclose(reg.yend[0], yend, RTOL)
        q = np.array([5.03])
        np.testing.assert_allclose(reg.q[0], q, RTOL)
        betas = np.array([[55.36575166], [0.46432416], [-0.66904404], [0.43205526]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([26.55390939])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        np.testing.assert_allclose(reg.e_filtered[0], np.array([31.74114306]), RTOL)
        predy = np.array([53.91309361])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        np.testing.assert_allclose(reg.n, 49, RTOL)
        np.testing.assert_allclose(reg.k, 3, RTOL)
        sig2 = 190.59435238060928
        np.testing.assert_allclose(reg.sig2, sig2)
        vm = np.array(
            [
                [5.52064057e02, -1.61264555e01, -8.86360735e00, 1.04251912e00],
                [-1.61264555e01, 5.44898242e-01, 2.39518645e-01, -1.88092950e-02],
                [-8.86360735e00, 2.39518645e-01, 1.55501840e-01, -2.18638648e-02],
                [1.04251912e00, -1.88092950e-02, -2.18638648e-02, 3.71222222e-02],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        its = 1
        np.testing.assert_allclose(reg.iteration, its, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        std_y = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, std_y)
        sig2 = 0
        # np.testing.assert_allclose(reg.sig2,sig2)
        hth = np.array(
            [
                [49.0, 704.371999, 139.75],
                [704.371999, 11686.67338121, 2246.12800625],
                [139.75, 2246.12800625, 498.5851],
            ]
        )
        np.testing.assert_allclose(reg.hth, hth, RTOL)


class GM_Endog_Error_Hom_Tester(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        self.X = np.array(X).T
        yd = []
        yd.append(db.by_col("CRIME"))
        self.yd = np.array(yd).T
        q = []
        q.append(db.by_col("DISCBD"))
        self.q = np.array(q).T
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = HOM.GM_Endog_Error_Hom(
            self.y, self.X, self.yd, self.q, self.w, A1="hom_sc"
        )
        np.testing.assert_allclose(reg.y[0], np.array([80.467003]), RTOL)
        x = np.array([1.0, 19.531])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        z = np.array([1.0, 19.531, 15.72598])
        np.testing.assert_allclose(reg.z[0], z, RTOL)
        h = np.array([1.0, 19.531, 5.03])
        np.testing.assert_allclose(reg.h[0], h, RTOL)
        yend = np.array([15.72598])
        np.testing.assert_allclose(reg.yend[0], yend, RTOL)
        q = np.array([5.03])
        np.testing.assert_allclose(reg.q[0], q, RTOL)
        betas = np.array([[55.36575166], [0.46432416], [-0.66904404], [0.43205526]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        u = np.array([26.55390939])
        np.testing.assert_allclose(reg.u[0], u, RTOL)
        np.testing.assert_allclose(reg.e_filtered[0], np.array([31.74114306]), RTOL)
        predy = np.array([53.91309361])
        np.testing.assert_allclose(reg.predy[0], predy, RTOL)
        np.testing.assert_allclose(reg.n, 49, RTOL)
        np.testing.assert_allclose(reg.k, 3, RTOL)
        vm = np.array(
            [
                [5.52064057e02, -1.61264555e01, -8.86360735e00, 1.04251912e00],
                [-1.61264555e01, 5.44898242e-01, 2.39518645e-01, -1.88092950e-02],
                [-8.86360735e00, 2.39518645e-01, 1.55501840e-01, -2.18638648e-02],
                [1.04251912e00, -1.88092950e-02, -2.18638648e-02, 3.71222222e-02],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        its = 1
        np.testing.assert_allclose(reg.iteration, its, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        std_y = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, std_y)
        pr2 = 0.34647366525657419
        np.testing.assert_allclose(reg.pr2, pr2)
        sig2 = 190.59435238060928
        np.testing.assert_allclose(reg.sig2, sig2)
        # std_err
        std_err = np.array([23.49604343, 0.73817223, 0.39433722, 0.19267128])
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        z_stat = np.array(
            [
                [2.35638617, 0.01845372],
                [0.62901874, 0.52933679],
                [-1.69662923, 0.08976678],
                [2.24244556, 0.02493259],
            ]
        )
        np.testing.assert_allclose(reg.z_stat, z_stat, RTOL)


class BaseGM_Combo_Hom_Tester(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        self.X = np.array(X).T
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        yd2, q2 = spreg.utils.set_endog(self.y, self.X, self.w, None, None, 1, True)
        self.X = np.hstack((np.ones(self.y.shape), self.X))
        reg = HOM.BaseGM_Combo_Hom(
            self.y, self.X, yend=yd2, q=q2, w=self.w.sparse, A1="hom_sc"
        )
        np.testing.assert_allclose(reg.y[0], np.array([80.467003]), RTOL)
        x = np.array([1.0, 19.531])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        betas = np.array([[10.12541428], [1.56832263], [0.15132076], [0.21033397]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        np.testing.assert_allclose(reg.u[0], np.array([34.3450723]), RTOL)
        np.testing.assert_allclose(reg.e_filtered[0], np.array([36.6149682]), RTOL)
        np.testing.assert_allclose(reg.predy[0], np.array([46.1219307]), RTOL)
        np.testing.assert_allclose(reg.n, 49, RTOL)
        np.testing.assert_allclose(reg.k, 3, RTOL)
        vm = np.array(
            [
                [2.33694742e02, -6.66856869e-01, -5.58304254e00, 4.85488380e00],
                [-6.66856869e-01, 1.94241504e-01, -5.42327138e-02, 5.37225570e-02],
                [-5.58304254e00, -5.42327138e-02, 1.63860721e-01, -1.44425498e-01],
                [4.85488380e00, 5.37225570e-02, -1.44425498e-01, 1.78622255e-01],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)
        z = np.array([1.0, 19.531, 35.4585005])
        np.testing.assert_allclose(reg.z[0], z, RTOL)
        h = np.array([1.0, 19.531, 18.594])
        np.testing.assert_allclose(reg.h[0], h, RTOL)
        yend = np.array([35.4585005])
        np.testing.assert_allclose(reg.yend[0], yend, RTOL)
        q = np.array([18.594])
        np.testing.assert_allclose(reg.q[0], q, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        its = 1
        np.testing.assert_allclose(reg.iteration, its, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        std_y = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, std_y)
        sig2 = 232.22680651270042
        # np.testing.assert_allclose(reg.sig2,sig2)
        np.testing.assert_allclose(reg.sig2, sig2)
        hth = np.array(
            [
                [49.0, 704.371999, 724.7435916],
                [704.371999, 11686.67338121, 11092.519988],
                [724.7435916, 11092.519988, 11614.62257048],
            ]
        )
        np.testing.assert_allclose(reg.hth, hth, RTOL)


class GM_Combo_Hom_Tester(unittest.TestCase):
    def setUp(self):
        db = libpysal.io.open(libpysal.examples.get_path("columbus.dbf"), "r")
        y = np.array(db.by_col("HOVAL"))
        self.y = np.reshape(y, (49, 1))
        X = []
        X.append(db.by_col("INC"))
        self.X = np.array(X).T
        self.w = libpysal.weights.Rook.from_shapefile(
            libpysal.examples.get_path("columbus.shp")
        )
        self.w.transform = "r"

    def test_model(self):
        reg = HOM.GM_Combo_Hom(self.y, self.X, w=self.w, A1="hom_sc")
        np.testing.assert_allclose(reg.y[0], np.array([80.467003]), RTOL)
        x = np.array([1.0, 19.531])
        np.testing.assert_allclose(reg.x[0], x, RTOL)
        betas = np.array([[10.12541428], [1.56832263], [0.15132076], [0.21033397]])
        np.testing.assert_allclose(reg.betas, betas, RTOL)
        np.testing.assert_allclose(reg.u[0], np.array([34.3450723]), RTOL)
        np.testing.assert_allclose(reg.e_filtered[0], np.array([36.6149682]), RTOL)
        np.testing.assert_allclose(reg.e_pred[0], np.array([32.90372983]), RTOL)
        np.testing.assert_allclose(reg.predy[0], np.array([46.1219307]), RTOL)
        np.testing.assert_allclose(reg.predy_e[0], np.array([47.56327317]), RTOL)
        np.testing.assert_allclose(reg.n, 49, RTOL)
        np.testing.assert_allclose(reg.k, 3, RTOL)
        z = np.array([1.0, 19.531, 35.4585005])
        np.testing.assert_allclose(reg.z[0], z, RTOL)
        h = np.array([1.0, 19.531, 18.594])
        np.testing.assert_allclose(reg.h[0], h, RTOL)
        yend = np.array([35.4585005])
        np.testing.assert_allclose(reg.yend[0], yend, RTOL)
        q = np.array([18.594])
        np.testing.assert_allclose(reg.q[0], q, RTOL)
        i_s = "Maximum number of iterations reached."
        np.testing.assert_string_equal(reg.iter_stop, i_s)
        np.testing.assert_allclose(reg.iteration, 1, RTOL)
        my = 38.436224469387746
        np.testing.assert_allclose(reg.mean_y, my)
        std_y = 18.466069465206047
        np.testing.assert_allclose(reg.std_y, std_y)
        pr2 = 0.28379825632694394
        np.testing.assert_allclose(reg.pr2, pr2)
        pr2_e = 0.25082892555141506
        np.testing.assert_allclose(reg.pr2_e, pr2_e)
        sig2 = 232.22680651270042
        # np.testing.assert_allclose(reg.sig2, sig2)
        np.testing.assert_allclose(reg.sig2, sig2)
        std_err = np.array([15.28707761, 0.44072838, 0.40479714, 0.42263726])
        np.testing.assert_allclose(reg.std_err, std_err, RTOL)
        z_stat = np.array(
            [
                [6.62351206e-01, 5.07746167e-01],
                [3.55847888e00, 3.73008780e-04],
                [3.73818749e-01, 7.08539170e-01],
                [4.97670189e-01, 6.18716523e-01],
            ]
        )
        np.testing.assert_allclose(reg.z_stat, z_stat, RTOL)
        vm = np.array(
            [
                [2.33694742e02, -6.66856869e-01, -5.58304254e00, 4.85488380e00],
                [-6.66856869e-01, 1.94241504e-01, -5.42327138e-02, 5.37225570e-02],
                [-5.58304254e00, -5.42327138e-02, 1.63860721e-01, -1.44425498e-01],
                [4.85488380e00, 5.37225570e-02, -1.44425498e-01, 1.78622255e-01],
            ]
        )
        np.testing.assert_allclose(reg.vm, vm, RTOL)


suite = unittest.TestSuite()
test_classes = [
    BaseGM_Error_Hom_Tester,
    GM_Error_Hom_Tester,
    BaseGM_Endog_Error_Hom_Tester,
    GM_Endog_Error_Hom_Tester,
    BaseGM_Combo_Hom_Tester,
    GM_Combo_Hom_Tester,
]
for i in test_classes:
    a = unittest.TestLoader().loadTestsFromTestCase(i)
    suite.addTest(a)

if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite)
